#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import copy
import traceback
from enum import Enum
from contextlib import contextmanager
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from cengal.code_flow_control.smart_values.versions.v_1.smart_values import *

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class ChainLinkFailed(Exception): pass


class ChainClosed(Exception):
    """
    link considers it ordinary external exception
    """

    def __init__(self, chain, link_id, link_info):
        super(ChainClosed, self).__init__('Content Holder is closed. '
                                          'Content Holder ID: {}; '
                                          'Content Holder Info: {}; '
                                          'Link ID: {}; '
                                          'Link Info {}'.format(str(chain._chain_id),
                                                                 str(chain._chain_info),
                                                                 str(link_id),
                                                                 str(link_info)))
        self.chain = chain
        self.link_id = link_id
        self.link_info = link_info


class ChainFunctionParameterNeeded(Exception): pass


class CriteriaType(Enum):
    # class CriteriaType():  # much more efficient than Enum inheritance
    needed = 0  # only set of this links is needed (should be already successfully done)
    optional = 1  # all links are needed except of this set of links
    any = 2  # any result will fit criteria
    forbidden = 3  # set of this links should be already failed
    not_successful = 4  # set of this links should not be successfully done (also may not start) at the check time


class IgnoreLinkResultCriteriaType(Enum):
    do_not_ignore = 0
    ignore_if_failed = 1
    ignore_if_successful = 2


class ChainHistoryExport(Exception):
    def __init__(self, history, process_error_result=True):
        super(ChainHistoryExport, self).__init__('')
        self.history = history
        self.process_error_result = process_error_result


class ChainInternalResult:
    def __init__(self, type_id, str_data, data):
        self.type_id = type_id
        self.str_data = str_data
        self.data = data

    def __str__(self):
        return self.str_data


class ChainInternalResultType(Enum):
    # class CriteriaType():  # much more efficient than Enum inheritance
    built_in_exception__chain_link_failed = 0
    built_in_exception__bad_history_import = 1
    external_exception = 2
    link_did_not_returned_an_answer = 3


class Chain:
    def __init__(self, chain_id=None, chain_info=None, global_link_results_criteria=None,
                 raise_exceptions=False, save_debug_trace=False, closeable=True):
        """

        :param chain_id:
        :param chain_info:
        :param global_link_results_criteria: will be set to ValueType(CriteriaType.optional, set()) if None;
            in this case all links are required.
        :param raise_exceptions:
        :param save_debug_trace:
        :param closeable:
        :return:
        """
        # Use only ValueType(CriteriaType.optional, ...) or ValueType(CriteriaType.needed, set()).
        # Other will be ignored here.
        # You may use global_link_results_criteria=ValueType(CriteriaType.optional, set()) to create criteria
        # "no fails in any link"
        self._chain_id = chain_id
        self._chain_info = chain_info
        self._internal_links_index = IDGenerator()
        self._reserve_link_id_generator = IDGenerator(GeneratorType.guid_string)
        self._criteria_list = list()
        global_link_results_criteria = global_link_results_criteria or ValueType(CriteriaType.optional, set())
        if global_link_results_criteria is not None:
            self._criteria_list.append(global_link_results_criteria)
        self._raise_exceptions = raise_exceptions
        self._save_debug_trace = save_debug_trace
        self._closeable = closeable
        self._full_history = list()
        self._links_library = dict()
        self._all_made_links = set()
        self._good_links = set()
        self._bad_links = set()

        self._current_link_id = None
        self._current_link_info = None
        self._current_link_result = None
        self._closed = False

        self._bool_result = ValueCache()

    # def _push_criteria(self, set_of_needed_links=None, set_of_optional_links=None):
    def _push_criteria(self, link_results_criteria):
        # Do not use!
        self._bool_result()
        self._criteria_list.append(link_results_criteria)

    def _pop_criteria(self):
        # Do not use!
        # May raise exception if len(self.criteria_list)==0, but this is OK.
        self._bool_result()
        return self._criteria_list.pop()

    def read_criteria(self):
        criteria = None
        if self._criteria_list:
            criteria = self._criteria_list[-1]
        return criteria

    def _push_link_info(self, link_id, link_info=None):
        self._current_link_id = link_id
        self._current_link_info = link_info
        self._current_link_result = None

    def _pop_link_info(self):
        self._current_link_id = None
        self._current_link_info = None
        self._current_link_result = None

    def push_result(self, bool_result, info_or_data=None):
        self._current_link_result = (bool_result, info_or_data)

    def push_result_c(self, result):
        # "class" version: to use when result = ValueExistence()
        self._current_link_result = (result.existence, result.result)

    def read_link_result_link(self, link_id):
        # result is NOT protected from changing!
        original_result_data = self._links_library[link_id][3]
        result = ValueExistence(original_result_data[0], original_result_data[1])
        # result = self._links_library[link_id][3]
        return result

    def read_link_result_copy(self, link_id):
        original_result_data = self._links_library[link_id][3]
        result = ValueExistence(original_result_data[0], copy.copy(original_result_data[1]))
        return result

    def read_link_result_deepcopy(self, link_id):
        original_result_data = self._links_library[link_id][3]
        result = ValueExistence(original_result_data[0], copy.deepcopy(original_result_data[1]))
        return result

    def _save_link_result(self, ignore_link_result_criteria=None):
        # ignore_link_result_criteria = ignore_link_result_criteria or IgnoreLinkResultCriteriaType.do_not_ignore
        if ((IgnoreLinkResultCriteriaType.ignore_if_failed == ignore_link_result_criteria) and
            (not self._current_link_result[0])) \
                or ((IgnoreLinkResultCriteriaType.ignore_if_successful == ignore_link_result_criteria) and
                    self._current_link_result[0]):
            return

        self._bool_result()
        import_depth = 0
        full_link_info = (self._internal_links_index(), self._current_link_id, self._current_link_info,
                           self._current_link_result, import_depth)
        self._full_history.append(full_link_info)
        self._links_library[self._current_link_id] = full_link_info
        self._all_made_links.add(self._current_link_id)
        if self._current_link_result[0]:
            self._good_links.add(self._current_link_id)
        else:
            self._bad_links.add(self._current_link_id)

    def __bool__(self):
        if self._bool_result:
            return self._bool_result.get()
        else:
            current_criteria = self.read_criteria()
            result = True
            if CriteriaType.needed == current_criteria:
                if len(current_criteria.result) != len(current_criteria.result & self._good_links):
                    result = False
            elif CriteriaType.optional == current_criteria:
                if len(self._bad_links - current_criteria.result) != 0:
                    result = False
            elif CriteriaType.any == current_criteria:
                result = True
            elif CriteriaType.forbidden == current_criteria:
                if len(current_criteria.result) != len(current_criteria.result & self._bad_links):
                    result = False
            elif CriteriaType.not_successful == current_criteria:
                if len(current_criteria.result & self._good_links) != 0:
                    result = False
            self._bool_result.set(result)
            return result

    def __nonzero__(self):
        return self.__bool__()

    @staticmethod
    def _link_list_to_str(link_list):
        links_str = ',\n'.join('(index({}), depth({}), ID({}), INFO({}), RESULT({}))'.format(str(another_link[0]),
                                                                                              str(another_link[4]),
                                                                                              str(another_link[1]),
                                                                                              str(another_link[2]),
                                                                                              '({}, {})'.format(
                                                                                                  str(another_link[3][0]),
                                                                                                  str(another_link[3][1])))
                                for another_link in link_list)
        return links_str

    def _link_str_to_chain_str(self, links_str):
        full_string = '{{{{CONTEXT_HOLDER_ID({}): CONTEXT_HOLDER_INFO({})}}:[\n{}\n]}}'.format(
                self._chain_id, self._chain_info, links_str)
        return full_string

    def get_bad_links(self):
        result = list()
        for another_link in self._full_history:
            if not another_link[3][0]:
                result.append(another_link)
        return result

    def get_bad_links_str(self):
        bad_links = self.get_bad_links()
        full_history_str = self._link_list_to_str(bad_links)
        full_string = self._link_str_to_chain_str(full_history_str)
        return full_string

    def raise_bad_links(self):
        raise ChainHistoryExport(self.get_bad_links())

    def raise_full_history(self):
        raise ChainHistoryExport(self._full_history)

    def process_history_import(self, his_ex):
        history = his_ex.history
        for another_link in history:
            full_link_info = (self._internal_links_index(), another_link[1], another_link[2],
                               another_link[3], another_link[4] + 1)
            self._full_history.append(full_link_info)

    def close(self):
        self._closed = True

    def _reopen(self):
        self._closed = False

    def __str__(self):
        full_history_str = self._link_list_to_str(self._full_history)
        full_string = self._link_str_to_chain_str(full_history_str)
        return full_string

    def __call__(self, *args, **kwargs):
        return link(self, *args, **kwargs)

    def chain(self, *args, **kwargs):
        return self.__call__(*args, **kwargs)


@contextmanager
def link(chain, link_id, link_info=None, link_results_criteria=None, ignore_link_result_criteria=None):
    if link_id is None:
        new_id = chain._reserve_link_id_generator()
        link_id = (new_id, new_id)

    if chain._closeable and chain._closed:
        raise ChainClosed(chain, link_id, link_info)

    chain._push_link_info(link_id, link_info)
    if link_results_criteria is not None:
        chain._push_criteria(link_results_criteria)
    need_to_save_link_result = True
    try:
        yield chain
    except ChainLinkFailed as exc:
        result_info = None
        if exc.args:
            result_info = exc.args[0]
        chain.push_result(False, ChainInternalResult(
                ChainInternalResultType.built_in_exception__chain_link_failed,
                'CHAIN INTERNAL RESULT. BUILT-IN EXCEPTION: ChainLinkFailed ({})'.format(result_info), result_info))
    except ChainHistoryExport as export:
        chain.process_history_import(export)
        if export.process_error_result:
            chain.push_result(False, ChainInternalResult(
                    ChainInternalResultType.built_in_exception__bad_history_import,
                    'CHAIN INTERNAL RESULT. BUILT-IN EXCEPTION: BAD HISTORY IMPORT EXCEPTION', None))
    except:
        exc = sys.exc_info()
        exc_type, exc_obj, exc_tb = exc
        tb_full_file_name = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        tb_line_number = exc_tb.tb_lineno
        tb_function_name = str()
        tb_text = str()

        tb_list = traceback.extract_tb(exc_tb, 2)
        if len(tb_list) >= 2:
            actual_tb = tb_list[1]
            tb_full_file_name, tb_line_number, tb_function_name, tb_text = actual_tb

        exception = exc
        error_str = '{} {}'.format(str(exception[0]), str(exception[1].args[0]))
        # print('+++', error_str)
        formatted_traceback = traceback.format_exception(exception[0], exception[1], exception[2])
        exception = exception[:2] + (formatted_traceback,)
        trace_str = ''.join(exception[2])
        if chain._save_debug_trace:
            result_string = 'CHAIN INTERNAL RESULT. CODE EXCEPTION "{}" AT "{}":{} in {} WITH TRACE: \n' \
                            '{}\n' \
                            '~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~' \
                            ''.format(error_str, tb_full_file_name, tb_line_number, tb_function_name, trace_str)
        else:
            result_string = 'CHAIN INTERNAL RESULT. CODE EXCEPTION "{}" AT "{}":{} in {}'.format(
                    error_str, tb_full_file_name, tb_line_number, tb_function_name)
        chain.push_result(False, ChainInternalResult(
                ChainInternalResultType.external_exception, result_string, exc))
        # print(result_string)
        # _chain_reader_runner__chain.push_result(False, sys.exc_info()[1])
        if chain._raise_exceptions:
            need_to_save_link_result = False
            chain._save_link_result()
            chain.raise_bad_links()
            # raise
    else:
        if chain._current_link_result is None:
            # _chain_reader_runner__chain.push_result(True)
            chain.push_result(False, ChainInternalResult(
                    ChainInternalResultType.link_did_not_returned_an_answer,
                    'CHAIN INTERNAL RESULT. Link DID NOT RETURN RESULT', None))
    finally:
        if need_to_save_link_result:
            chain._save_link_result(ignore_link_result_criteria)
        if link_results_criteria is not None:
            chain._pop_criteria()
        chain._pop_link_info()


def link__function(target_function):
    """
    Parameters: chain__chain= (required)
        , chain__link_id= (required)
        , chain__link_info= (optional)
        , chain__link_results_criteria= (optional)
        , chain__ignore_link_result_criteria= (optional).
    Parameters passed to the target_function: chain__chain (after local link configuration).
    :param target_function: function
    :return:
    """

    def new_target_function(*args, **kwargs):
        chain = None
        if 'chain__chain' in kwargs:
            chain = kwargs['chain__chain']
            del kwargs['chain__chain']
        else:
            raise ChainFunctionParameterNeeded('chain__chain')

        link_id = None
        if 'chain__link_id' in kwargs:
            link_id = kwargs['chain__link_id']
            del kwargs['chain__link_id']
        else:
            raise ChainFunctionParameterNeeded('chain__link_id')

        link_info = None
        if 'chain__link_info' in kwargs:
            link_info = kwargs['chain__link_info']
            del kwargs['chain__link_info']

        link_results_criteria = None
        if 'chain__link_results_criteria' in kwargs:
            link_results_criteria = kwargs['chain__link_results_criteria']
            del kwargs['chain__link_results_criteria']

        ignore_link_result_criteria = None
        if 'chain__ignore_link_result_criteria' in kwargs:
            ignore_link_result_criteria = kwargs['chain__ignore_link_result_criteria']
            del kwargs['chain__ignore_link_result_criteria']

        target_function_result = None
        with link(chain, link_id, link_info, link_results_criteria, ignore_link_result_criteria) as \
                context:
            kwargs['chain__chain'] = context
            target_function_result = target_function(*args, **kwargs)

        return target_function_result

    return new_target_function


class _ChainRunner:
    def __init__(self, current_globals, chain, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None, function_result_criteria=None):
        self._chain_runner__current_globals = current_globals
        self._chain_runner__chain = chain
        self._chain_runner__link_id = link_id
        self._chain_runner__link_info = link_info
        self._chain_runner__link_results_criteria = link_results_criteria
        self._chain_runner__ignore_link_result_criteria = ignore_link_result_criteria

    def __getattr__(self, name):
        target_functor = None
        if name in self._chain_runner__current_globals:
            target_functor = self._chain_runner__current_globals[name]
        else:
            raise AttributeError(name)

        def new_target_function(*args, **kwargs):
            target_function_result = None
            link_id = self._chain_runner__link_id or str(target_functor)
            with link(self._chain_runner__chain, link_id, self._chain_runner__link_info,
                      self._chain_runner__link_results_criteria,
                      self._chain_runner__ignore_link_result_criteria) as context:
                kwargs['chain__chain'] = context
                target_function_result = target_functor(*args, **kwargs)

            return target_function_result

        return new_target_function


class ChainRunner:
    def __init__(self, current_globals, chain):
        self.current_globals = current_globals
        self.chain = chain

    def __call__(self, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None):
        result = _ChainRunner(self.current_globals, self.chain, link_id, link_info, link_results_criteria,
                              ignore_link_result_criteria)
        return result


class _ChainCallRunner:
    def __init__(self, chain, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None, function_result_criteria=None):
        self._chain_runner__chain = chain
        self._chain_runner__link_id = link_id
        self._chain_runner__link_info = link_info
        self._chain_runner__link_results_criteria = link_results_criteria
        self._chain_runner__ignore_link_result_criteria = ignore_link_result_criteria

    def __call__(self, target_functor, *args, **kwargs):
        target_function_result = None
        link_id = self._chain_runner__link_id or str(target_functor)
        with link(self._chain_runner__chain, link_id, self._chain_runner__link_info,
                  self._chain_runner__link_results_criteria,
                  self._chain_runner__ignore_link_result_criteria) as context:
            kwargs['chain__chain'] = context
            target_function_result = target_functor(*args, **kwargs)

        return target_function_result


class ChainCallRunner:
    def __init__(self, chain):
        self.chain = chain

    def __call__(self, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None):
        result = _ChainCallRunner(self.chain, link_id, link_info, link_results_criteria,
                                  ignore_link_result_criteria)
        return result


def link__function__simple(target_function):
    """
    Parameters: chain__chain= (required)
        , chain__link_id= (optional) (default value == str(target_function))
        , chain__link_results_criteria= (optional)
        , chain__ignore_link_result_criteria= (optional).
    Parameters passed to the target_function: .
    :param target_function: function
    :return:
    """

    def new_target_function(*args, **kwargs):
        chain = None
        if 'chain__chain' in kwargs:
            chain = kwargs['chain__chain']
            del kwargs['chain__chain']
        else:
            raise ChainFunctionParameterNeeded('chain__chain')

        # link_id = '__UNNAMED_FUNCTION_SIMPLE_LINK__'
        link_id = str(target_function)
        if 'chain__link_id' in kwargs:
            link_id = kwargs['chain__link_id']
            del kwargs['chain__link_id']

        link_results_criteria = None
        if 'chain__link_results_criteria' in kwargs:
            link_results_criteria = kwargs['chain__link_results_criteria']
            del kwargs['chain__link_results_criteria']

        ignore_link_result_criteria = None
        if 'chain__ignore_link_result_criteria' in kwargs:
            ignore_link_result_criteria = kwargs['chain__ignore_link_result_criteria']
            del kwargs['chain__ignore_link_result_criteria']

        target_function_result = None
        with link(chain, link_id, None, link_results_criteria, ignore_link_result_criteria) as context:
            if context:
                target_function_result = target_function(*args, **kwargs)
                context.push_result(True, target_function_result)

        return target_function_result

    return new_target_function


class _ChainRunnerSimple:
    def __init__(self, current_globals, chain, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None, function_result_criteria=None):
        self._chain_runner_simple__current_globals = current_globals
        self._chain_runner_simple__chain = chain
        self._chain_runner_simple__link_id = link_id
        self._chain_runner_simple__link_info = link_info
        self._chain_runner_simple__link_results_criteria = link_results_criteria
        self._chain_runner_simple__ignore_link_result_criteria = ignore_link_result_criteria
        self._chain_runner_simple__function_result_criteria = function_result_criteria or (lambda result: True)

    def __getattr__(self, name):
        target_functor = None
        if name in self._chain_runner_simple__current_globals:
            target_functor = self._chain_runner_simple__current_globals[name]
        else:
            raise AttributeError(name)

        def new_target_function(*args, **kwargs):
            target_function_result = None
            link_id = self._chain_runner_simple__link_id or str(target_functor)
            with link(self._chain_runner_simple__chain, link_id, self._chain_runner_simple__link_info,
                      self._chain_runner_simple__link_results_criteria,
                      self._chain_runner_simple__ignore_link_result_criteria) as context:
                if context:
                    target_function_result = target_functor(*args, **kwargs)
                    is_good_result = self._chain_runner_simple__function_result_criteria(target_function_result)
                    context.push_result(is_good_result, target_function_result)

            return target_function_result

        return new_target_function


class ChainUniRunner:
    def __init__(self, current_globals, chain, simple_mode=False, function_result_criteria=None):
        self.current_globals = current_globals
        self.chain = chain
        self.simple_mode = simple_mode
        self.default_function_result_criteria = function_result_criteria or (lambda result: True)
        self.function_result_criteria = self.default_function_result_criteria

        self.runner_class = _ChainRunner
        if self.simple_mode:
            self.runner_class = _ChainRunnerSimple

    def set_function_result_criteria(self, result_criteria_computer):
        self.function_result_criteria = result_criteria_computer

    def reset_function_result_criteria(self):
        self.function_result_criteria = self.default_function_result_criteria

    def __call__(self, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None):
        result = self.runner_class(self.current_globals, self.chain, link_id, link_info,
                                   link_results_criteria, ignore_link_result_criteria, self.function_result_criteria)
        return result


class _ChainCallRunnerSimple:
    def __init__(self, chain, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None, function_result_criteria=None):
        self._chain_runner_simple__chain = chain
        self._chain_runner_simple__link_id = link_id
        self._chain_runner_simple__link_info = link_info
        self._chain_runner_simple__link_results_criteria = link_results_criteria
        self._chain_runner_simple__ignore_link_result_criteria = ignore_link_result_criteria
        self._chain_runner_simple__function_result_criteria = function_result_criteria or (lambda result: True)

    def __call__(self, target_functor, *args, **kwargs):
        target_function_result = None
        link_id = self._chain_runner_simple__link_id or str(target_functor)
        with link(self._chain_runner_simple__chain, link_id, self._chain_runner_simple__link_info,
                  self._chain_runner_simple__link_results_criteria,
                  self._chain_runner_simple__ignore_link_result_criteria) as context:
            if context:
                target_function_result = target_functor(*args, **kwargs)
                is_good_result = self._chain_runner_simple__function_result_criteria(target_function_result)
                context.push_result(is_good_result, target_function_result)

        return target_function_result


class ChainUniCallRunner:
    def __init__(self, chain, simple_mode=False, function_result_criteria=None):
        self.chain = chain
        self.simple_mode = simple_mode
        self.default_function_result_criteria = function_result_criteria or (lambda result: True)
        self.function_result_criteria = self.default_function_result_criteria

        self.runner_class = _ChainCallRunner
        if self.simple_mode:
            self.runner_class = _ChainCallRunnerSimple

    def set_function_result_criteria(self, result_criteria_computer):
        self.function_result_criteria = result_criteria_computer

    def reset_function_result_criteria(self):
        self.function_result_criteria = self.default_function_result_criteria

    def __call__(self, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None):
        result = self.runner_class(self.chain, link_id, link_info,
                                   link_results_criteria, ignore_link_result_criteria, self.function_result_criteria)
        return result


class _ChainValRunner:
    def __init__(self, chain, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None, function_result_criteria=None, reaction_to_the_result=None):
        self._chain_runner_simple__chain = chain
        self._chain_runner_simple__link_id = link_id
        self._chain_runner_simple__link_info = link_info
        self._chain_runner_simple__link_results_criteria = link_results_criteria
        self._chain_runner_simple__ignore_link_result_criteria = ignore_link_result_criteria
        self._chain_runner_simple__function_result_criteria = function_result_criteria or (lambda result: True)
        self._chain_runner_simple__reaction_to_the_result = reaction_to_the_result

    def __call__(self, functor_result):
        target_function_result = functor_result
        link_id = self._chain_runner_simple__link_id
        with link(self._chain_runner_simple__chain, link_id, self._chain_runner_simple__link_info,
                  self._chain_runner_simple__link_results_criteria,
                  self._chain_runner_simple__ignore_link_result_criteria) as context:
            if context:
                is_good_result = self._chain_runner_simple__function_result_criteria(target_function_result)
                if self._chain_runner_simple__reaction_to_the_result is not None:
                    verdict = self._chain_runner_simple__reaction_to_the_result(is_good_result, target_function_result)
                    target_function_result = (target_function_result, verdict)
                context.push_result(is_good_result, target_function_result)

        return functor_result


class ChainValRunner:
    def __init__(self, chain, function_result_criteria=None, reaction_to_the_result=None):
        self.chain = chain
        self.default_function_result_criteria = function_result_criteria or (lambda result: True)
        self.function_result_criteria = self.default_function_result_criteria
        self.reaction_to_the_result = reaction_to_the_result

    def set_function_result_criteria(self, result_criteria_computer):
        self.function_result_criteria = result_criteria_computer

    def reset_function_result_criteria(self):
        self.function_result_criteria = self.default_function_result_criteria

    def __call__(self, link_id=None, link_info=None, link_results_criteria=None,
                 ignore_link_result_criteria=None):
        result = _ChainValRunner(self.chain, link_id, link_info,
                                 link_results_criteria, ignore_link_result_criteria,
                                 self.function_result_criteria, self.reaction_to_the_result)
        return result


@contextmanager
def chain_reader(chain, link_results_criteria=None, close=False):
    if link_results_criteria is not None:
        chain._push_criteria(link_results_criteria)
    try:
        yield chain
    except:
        raise
    finally:
        if link_results_criteria is not None:
            chain._pop_criteria()
        if close:
            chain.close()


def chain_reader__function(target_function):
    """
    Parameters: chain__chain= (required), chain__link_results_criteria= (optional), chain__close= (optional).
    Parameters passed to the target_function: chain__chain (after local link configuration).
    :param target_function: function
    :return:
    """

    def new_target_function(*args, **kwargs):
        chain = None
        if 'chain__chain' in kwargs:
            chain = kwargs['chain__chain']
            del kwargs['chain__chain']
        else:
            raise ChainFunctionParameterNeeded('chain__chain')

        link_results_criteria = None
        if 'chain__link_results_criteria' in kwargs:
            link_results_criteria = kwargs['chain__link_results_criteria']
            del kwargs['chain__link_results_criteria']

        close = None
        if 'chain__close' in kwargs:
            close = kwargs['chain__close']
            del kwargs['chain__close']

        target_function_result = None
        with chain_reader(chain, link_results_criteria, close) as context:
            kwargs['chain__chain'] = context
            target_function_result = target_function(*args, **kwargs)

        return target_function_result

    return new_target_function


class _ChainReaderRunner:
    def __init__(self, current_globals, chain, link_results_criteria=None, close=False):
        self._chain_reader_runner__current_globals = current_globals
        self._chain_reader_runner__chain = chain
        self._chain_reader_runner__link_results_criteria = link_results_criteria
        self._chain_reader_runner__close = close

    def __getattr__(self, name):
        target_functor = None
        if name in self._chain_reader_runner__current_globals:
            target_functor = self._chain_reader_runner__current_globals[name]
        else:
            raise AttributeError(name)

        def new_target_function(*args, **kwargs):
            target_function_result = None
            with chain_reader(self._chain_reader_runner__chain,
                              self._chain_reader_runner__link_results_criteria,
                              self._chain_reader_runner__close) as context:
                kwargs['chain__chain'] = context
                target_function_result = target_functor(*args, **kwargs)

            return target_function_result

        return new_target_function


class ChainReaderRunner:
    def __init__(self, current_globals, chain):
        self.current_globals = current_globals
        self.chain = chain

    def __call__(self, link_results_criteria=None, close=False):
        result = _ChainReaderRunner(self.current_globals, self.chain, link_results_criteria, close)
        return result


class _ChainReaderCallRunner:
    def __init__(self, chain, link_results_criteria=None, close=False):
        self._chain_reader_runner__chain = chain
        self._chain_reader_runner__link_results_criteria = link_results_criteria
        self._chain_reader_runner__close = close

    def __call__(self, target_functor, *args, **kwargs):
        target_function_result = None
        with chain_reader(self._chain_reader_runner__chain,
                          self._chain_reader_runner__link_results_criteria,
                          self._chain_reader_runner__close) as context:
            kwargs['chain__chain'] = context
            target_function_result = target_functor(*args, **kwargs)

        return target_function_result


class ChainReaderCallRunner:
    def __init__(self, chain):
        self.chain = chain

    def __call__(self, link_results_criteria=None, close=False):
        result = _ChainReaderCallRunner(self.chain, link_results_criteria, close)
        return result



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


IF (PYXD_UNAME_SYSNAME == "Windows") and (PYXD_UNAME_MACHINE in ("x86_64", "x86", "i386", "i686", "AMD64")):
    cdef extern from "<immintrin.h>" nogil:
        void _mm_pause()


    cpdef void mm_pause():
        _mm_pause()


ELIF (PYXD_UNAME_SYSNAME in ("Linux", "Darwin")) and (PYXD_UNAME_MACHINE in ("x86_64", "x86", "i386", "i686", "AMD64")):
    cdef extern from "<immintrin.h>" nogil:
        void _mm_pause()


    cpdef void mm_pause():
        _mm_pause()


ELSE:
    cpdef void mm_pause():
        pass


IF PYXD_UNAME_SYSNAME == "Windows":
    from libc.time cimport clock, CLOCKS_PER_SEC, clock_t

    cdef extern from "Windows.h":
        int SwitchToThread() nogil


    cpdef void cython_spinwait(double sleep_time):
        cdef clock_t start, end

        start = clock()
        end = start + <clock_t>(sleep_time * CLOCKS_PER_SEC)

        while clock() < end:
            mm_pause()
            SwitchToThread()
            mm_pause()
    

    hps_sleep = high_precision_sync_sleep = cython_spinwait

ELIF PYXD_UNAME_SYSNAME in ("Linux", "Darwin"):
    cdef extern from "time.h":
        struct timespec:
            long tv_sec
            long tv_nsec


    cdef extern from "unistd.h":
        int nanosleep(const timespec *req, timespec *rem) nogil


    cpdef void cython_nanosleep(double sleep_time):
        cdef timespec sleep_req
        cdef timespec sleep_rem

        sleep_req.tv_sec = <long>(sleep_time)
        sleep_req.tv_nsec = <long>((sleep_time - sleep_req.tv_sec) * 1e9)

        while nanosleep(&sleep_req, &sleep_rem) == -1:
            mm_pause()
            sleep_req = sleep_rem
            mm_pause()
    
    hps_sleep = high_precision_sync_sleep = cython_nanosleep

ELIF PYXD_UNAME_SYSNAME == "Emscripten":
    from libc.stdint cimport uint32_t

    cdef extern from "emscripten.h":
        void emscripten_sleep(uint32_t ms) nogil

    cpdef void cython_emscripten_sleep(double sleep_time):
        cdef uint32_t sleep_ms

        sleep_ms = <uint32_t>(sleep_time * 1e3)

        emscripten_sleep(sleep_ms)
    
    
    hps_sleep = high_precision_sync_sleep = cython_emscripten_sleep



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import os
import errno


if 'nt' == os.name:
    if not hasattr(errno, 'WSAEPIPE'):
        WSAEPIPE = -12345
    
    SET_OF_CONNECTION_ERRORS = {errno.WSAECONNRESET, errno.WSAECONNREFUSED, errno.WSAECONNABORTED, WSAEPIPE, errno.WSAESHUTDOWN}
    SET_OF_UDP_MESSAGE_SIZE_RELATED_ERRORS = {errno.WSAEMSGSIZE}
else:
    SET_OF_CONNECTION_ERRORS = {errno.ECONNRESET, errno.ECONNREFUSED, errno.ECONNABORTED, errno.EPIPE, errno.ESHUTDOWN}
    SET_OF_UDP_MESSAGE_SIZE_RELATED_ERRORS = {errno.EMSGSIZE}



try:
    BlockingIOError
except NameError:
    class BlockingIOError(OSError):
        pass

try:
    InterruptedError
except NameError:
    class InterruptedError(OSError):
        pass

try:
    ConnectionError
except NameError:
    class ConnectionError(OSError):
        pass

try:
    BrokenPipeError
except NameError:
    class BrokenPipeError(ConnectionError):
        pass

try:
    ConnectionAbortedError
except NameError:
    class ConnectionAbortedError(ConnectionError):
        pass

try:
    ConnectionRefusedError
except NameError:
    class ConnectionRefusedError(ConnectionError):
        pass

try:
    ConnectionResetError
except NameError:
    class ConnectionResetError(ConnectionError):
        pass



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from enum import Enum
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.asyncio_loop import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import *
from cengal.parallel_execution.coroutines.coro_standard_services.timer_func_runner import *
from cengal.file_system.file_manager import path_relative_to_current_dir
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.data_manipulation.serialization import *
from typing import Hashable, Tuple, List, Any, Dict, Callable, Type
from cengal.introspection.inspect import get_exception, entity_owning_module_importable_str, entity_owning_module_info_and_owning_path, entity_properties
from cengal.io.core.memory_management import IOCoreMemoryManagement
from cengal.parallel_execution.asyncio.efficient_streams import StreamManagerIOCoreMemoryManagement, TcpStreamManager, UdpStreamManager, StreamManagerAbstract
from cengal.code_flow_control.smart_values import ValueExistence
from cengal.io.named_connections.named_connections_manager import NamedConnectionsManager
from cengal.code_flow_control.args_manager import number_of_provided_args
from cengal.data_manipulation.serialization import Serializer, Serializers, best_serializer_for_standard_data
from cengal.code_flow_control.args_manager import find_arg_position_and_value, UnknownArgumentError
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from cengal.system import PLATFORM_NAME, PYTHON_VERSION
from importlib import import_module
import sys
import os
import asyncio

from .exceptions import *
from .commands import *
from .class_info import *
from .request_class_info import *
from .serializers import *


class RemoteNode:
    def __init__(self) -> None:
        self.foreign_platform_name: str = None
        self.serializers: Dict[SerializerID, Serializer] = dict()
        self.request_id_gen: IDGenerator = IDGenerator(GeneratorType.integer)
    
    def register_node(self, data: Dict):
        raise NotImplementedError
    
    def gen_request_id(self) -> int:
        return self.request_id_gen()

    def register_service_class(self, *args, **kwargs):
        raise NotImplementedError

    def register_request_class(self, *args, **kwargs):
        raise NotImplementedError


class RemoteServer(RemoteNode):
    def __init__(self) -> None:
        super().__init__()
        self.service_class_id_gen: IDGenerator = IDGenerator(GeneratorType.integer)
        self.remotely_registered_service_classes: Dict[Type, LocalClassInfo] = dict()
        
        self.request_class_id_gen: IDGenerator = IDGenerator(GeneratorType.integer)
        self.remotely_registered_request_classes: Dict[Type, LocalRequestClassInfo] = dict()
    
    def register_service_class(self, class_type: Type[Service]):
        self.remotely_registered_service_classes[class_type] = LocalClassInfo(self.service_class_id_gen(), class_type)
    
    def register_request_class(self, request: Request):
        self.remotely_registered_request_classes[type(request)] = LocalRequestClassInfo(self.request_class_id_gen(), request)


class RemoteClient(RemoteNode):
    def __init__(self) -> None:
        super().__init__()
        self.locally_registered_service_classes: Dict[Hashable, RemoteClassInfo] = dict()
        self.locally_registered_request_classes: Dict[Hashable, RemoteClassInfo] = dict()
    
    def register_service_class(self, data: Dict):
        remote_class_info: RemoteClassInfo = RemoteClassInfo.from_data(data)
        self.locally_registered_service_classes[remote_class_info.local_id] = remote_class_info
    
    def register_request_class(self, data: Dict):
        remote_request_class_info: RemoteRequestClassInfo = RemoteRequestClassInfo.from_data(data)
        self.locally_registered_request_classes[remote_request_class_info.local_id] = remote_request_class_info



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['ensure_copyright', 'get_version', 'set_version', 'default_version_bracket_pair']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.file_system.file_manager import current_src_file_dir, path_relative_to_current_src
from cengal.file_system.directory_manager import filtered_file_list_traversal, FilteringType
from cengal.file_system.path_manager import path_relative_to_src, RelativePath
from cengal.text_processing.encoding_detection import detect_and_decode
from cengal.text_processing.text_processing import find_text, replace_slice
from cengal.text_processing.brackets_processing import BracketPair, Bracket, replace_text_with_brackets, find_text_with_brackets, find_text_in_brackets, replace_text_in_brackets
from os.path import splitext, join
from typing import Set, Optional


_PYTHON_FILES_EXTENSIONS: Set[str] = {'.pyx', '.py', '.pyw', 'pyi'}


def ensure_copyright(src_root_dir: str, 
                     head_string: str, 
                     license_string: str, 
                     license_bracket_pair: BracketPair, 
                     module_docstring_string: str, 
                     credits_string: str,
                     credits_bracket_pair: BracketPair,
                     python_files_extensions: Set[str] = None,
                     silent: bool = False):
    python_files_extensions = python_files_extensions or _PYTHON_FILES_EXTENSIONS
    travers_result = filtered_file_list_traversal(src_root_dir, FilteringType.including, _PYTHON_FILES_EXTENSIONS, remove_empty_items=True, use_spinner=False)
    files_processed: int = 0
    try:
        for dir_path, dirs, files in travers_result:
            for file in files:
                file_path = join(dir_path, file)
                if not silent:
                    print()
                    print(f'... Processing: {file_path}')

                try:
                    with open(file_path, 'r+b') as file:
                        text, text_encoding, bom_bytes = detect_and_decode(file.read())
                        if not bom_bytes:
                            text_encoding = 'utf-8'
                        
                        head_string_pos = find_text(text, head_string, 0, len(head_string) * 2)
                        license_string_pos = find_text_with_brackets(text, license_bracket_pair)
                        module_docstring_string_pos = find_text(text, module_docstring_string)
                        credits_string_pos = find_text_with_brackets(text, credits_bracket_pair)
                        
                        replace_only: bool = bool(head_string_pos is not None) and bool(license_string_pos is not None) and bool(credits_string_pos is not None)
                        if replace_only:
                            text, _ = replace_text_with_brackets(text, license_bracket_pair, license_string, 1)
                            text, _ = replace_text_with_brackets(text, credits_bracket_pair, credits_string, 1)
                        else:
                            head_string_pos = find_text(text, head_string, 0, len(head_string) * 2)
                            if head_string_pos is not None:
                                text, _ = replace_slice(text, head_string_pos, str())
                            
                            license_string_pos = find_text_with_brackets(text, license_bracket_pair)
                            if license_string_pos is not None:
                                text, _ = replace_slice(text, license_string_pos, str())
                            
                            module_docstring_string_pos = find_text(text, module_docstring_string)
                            if module_docstring_string_pos is not None:
                                text, _ = replace_slice(text, module_docstring_string_pos, str())
                            
                            credits_string_pos = find_text_with_brackets(text, credits_bracket_pair)
                            if credits_string_pos is not None:
                                text, _ = replace_slice(text, credits_string_pos, str())

                            new_head: str = head_string + '\n\n' + license_string + '\n\n\n' + module_docstring_string + '\n\n\n' + credits_string + '\n\n\n'
                            text = new_head + text
                        
                        file.seek(0, 0)
                        file.truncate(0)
                        data = bom_bytes + text.encode(text_encoding)
                        file.write(data)
                        files_processed += 1
                finally:
                    if not silent:
                        print(f'Done: {file_path}')
    finally:
        if not silent:
            print()
            print(f'Files procesed: {files_processed}')


default_credits_bracket_pair: BracketPair = BracketPair([Bracket('credits_string =')], [Bracket('credits_bracket_pair: BracketPair =')])
default_version_bracket_pair: BracketPair = BracketPair([Bracket('__version__ = "')], [Bracket('"')])


def get_version(script_file_full_path: str,
                credits_string_bracket_pair: Optional[BracketPair] = None, 
                version_bracket_pair: Optional[BracketPair] = None, 
                ) -> Optional[str]:
    credits_string_bracket_pair = default_credits_bracket_pair if credits_string_bracket_pair is None else credits_string_bracket_pair
    version_bracket_pair = default_version_bracket_pair if version_bracket_pair is None else version_bracket_pair
    with open(script_file_full_path, 'rb') as file:
        text, _, _ = detect_and_decode(file.read())
        credits_string_pos: slice = find_text_in_brackets(text, credits_string_bracket_pair)
        if credits_string_pos is not None:
            credits_string: str = text[credits_string_pos]
            version_pos: slice = find_text_in_brackets(credits_string, version_bracket_pair)
            if version_pos is not None:
                return credits_string[version_pos]
    
    return None


def set_version(script_file_full_path: str,
                new_version: str, 
                credits_string_bracket_pair: Optional[BracketPair] = None, 
                version_bracket_pair: Optional[BracketPair] = None, 
                ) -> bool:
    credits_string_bracket_pair = default_credits_bracket_pair if credits_string_bracket_pair is None else credits_string_bracket_pair
    version_bracket_pair = default_version_bracket_pair if version_bracket_pair is None else version_bracket_pair
    with open(script_file_full_path, 'r+b') as file:
        text, text_encoding, bom_bytes = detect_and_decode(file.read())
        credits_string_pos: slice = find_text_in_brackets(text, credits_string_bracket_pair)
        if credits_string_pos is not None:
            credits_string: str = text[credits_string_pos]
            version_pos: slice = find_text_in_brackets(credits_string, version_bracket_pair)
            if version_pos is not None:
                credits_string, _ = replace_text_in_brackets(credits_string, version_bracket_pair, new_version)
                text, _ = replace_slice(text, credits_string_pos, credits_string)
                file.seek(0, 0)
                file.truncate(0)
                data = bom_bytes + text.encode(text_encoding)
                file.write(data)
                return True
    
    return False



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import clock

perf_counter = process_time = clock

try:
    from time import perf_counter as _perf_counter
    perf_counter = _perf_counter
except ImportError:
    pass

try:
    from time import process_time as _process_time
    process_time = _process_time
except ImportError:
    pass

from cengal.time_management.repeat_for_a_time.versions.v_0.repeat_for_a_time__python import \
    Tracer as TracerNative, TracerCounter as TracerCounterNative, TracerIterator as TracerIteratorNative, \
    ClockType as ClockTypeNative

repeat_for_a_time_cython_module_was_loaded = True
try:
    from cengal.time_management.repeat_for_a_time.versions import \
        Tracer as TracerCompiled, TracerCounter as TracerCounterCompiled, TracerIterator as TracerIteratorCompiled, \
        ClockType as ClockTypeCompiled
except ImportError:
    repeat_for_a_time_cython_module_was_loaded = False

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"

__doc__ = '''
Native .iter(): 1314866.836082816 iter/s; 1.0101098936807824 seconds. Internal: 1267625.7394312301 iter/s; 1.0100524425506592 seconds; 1328160 iters
Native Counter: 1374039.594745121 iter/s; 0.9225542006561607 seconds. Internal: 1373622.4018055892 iter/s; 0.9228343963623047 seconds; 1267626 iters
Native Counter Iterator: 833126.4583887773 iter/s; 1.5215289194530226 seconds. Internal: 833211.1709637797 iter/s; 1.521374225616455 seconds; 1267626 iters
PASS - Native .iter(): 2679790.563039528 iter/s; 0.9975746003702035 seconds. Internal: 2704377.710817054 iter/s; 0.9975225925445557 seconds; 2673291 iters
PASS - Native Counter: 2573459.392848136 iter/s; 1.050872614316627 seconds. Internal: 2573991.52006913 iter/s; 1.0506553649902344 seconds; 2704378 iters
PASS - Native Counter Iterator: 1358141.8445628048 iter/s; 1.991233839695557 seconds. Internal: 1357936.3387444955 iter/s; 1.9915351867675781 seconds; 2704378 iters

Compiled .iter(): 2662605.4954643254 iter/s; 0.9982654225448746 seconds. Internal: 2670643.0 iter/s; 0.9990000128746033 seconds; 2657987 iters
Compiled Counter: 2552670.609531415 iter/s; 1.0462152813716301 seconds. Internal: 2553196.0 iter/s; 1.0460000038146973 seconds; 2670643 iters
Compiled Counter Iterator: 2412906.0912160017 iter/s; 1.1068159717123969 seconds. Internal: 2412505.0 iter/s; 1.1069999933242798 seconds; 2670643 iters
PASS - Compiled .iter(): 10369459.449145418 iter/s; 1.0086939489273004 seconds. Internal: 10004845.0 iter/s; 1.0089999437332153 seconds; 10459611 iters
PASS - Compiled Counter: 10717891.79297866 iter/s; 0.9334713573572575 seconds. Internal: 10723306.0 iter/s; 0.9330000281333923 seconds; 10004845 iters
PASS - Compiled Counter Iterator: 8438850.11835808 iter/s; 1.1855696996247413 seconds. Internal: 8435788.0 iter/s; 1.1859999895095825 seconds; 10004845 iters

for i in range(): 3257425.63035198 iter/s
PASS - for i in range(): 27120025.0597179 iter/s

while iter_qnt > 0: 2441599.852625212 iter/s
PASS - while iter_qnt > 0: 9778029.90225164 iter/s

while True: 2592510.421169292 iter/s
PASS - while True: 9770496.61187559 iter/s
'''

test_time = 1.0

# =============================================================
clock_type_for_native = ClockTypeNative.perf_counter
clock_type_for_native_tracer = clock_type_for_native
clock_type_for_native_tracer_counter = ClockTypeNative.fake

tr = TracerNative(test_time, clock_type_for_native_tracer)
start_time = perf_counter()
while tr.iter():
    k = int('1243')
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = tr.iterations_made / time_spent
print('Native .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
start_time = perf_counter()
while trc.iter():
    k = int('1243')
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = trc.iterations_made / time_spent
print('Native Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
start_time = perf_counter()
for i in TracerIteratorNative(trc):
    k = int('1243')
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = trc.iterations_made / time_spent
print('Native Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

tr = TracerNative(test_time, clock_type_for_native_tracer)
start_time = perf_counter()
while tr.iter():
    pass
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = tr.iterations_made / time_spent
print('PASS - Native .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
start_time = perf_counter()
while trc.iter():
    pass
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = trc.iterations_made / time_spent
print('PASS - Native Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
start_time = perf_counter()
for i in TracerIteratorNative(trc):
    pass
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = trc.iterations_made / time_spent
print('PASS - Native Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

print()

# =============================================================

if repeat_for_a_time_cython_module_was_loaded:
    clock_type_for_compiled = ClockTypeCompiled.perf_counter
    clock_type_for_compiled_tracer = clock_type_for_compiled
    clock_type_for_compiled_tracer_counter = ClockTypeCompiled.fake

    tr = TracerCompiled(test_time, clock_type_for_compiled_tracer)
    start_time = perf_counter()
    while tr.iter():
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = tr.iterations_made / time_spent
    print('Compiled .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    trc = TracerCounterCompiled(tr.iter_per_time_unit, test_time, clock_type_for_compiled_tracer_counter)
    start_time = perf_counter()
    while trc.iter():
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('Compiled Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    trc = TracerCounterCompiled(tr.iter_per_time_unit, test_time, clock_type_for_compiled_tracer_counter)
    start_time = perf_counter()
    for i in TracerIteratorCompiled(trc):
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('Compiled Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    tr = TracerCompiled(test_time, clock_type_for_compiled_tracer)
    start_time = perf_counter()
    while tr.iter():
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = tr.iterations_made / time_spent
    print('PASS - Compiled .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    trc = TracerCounterCompiled(tr.iter_per_time_unit, test_time, clock_type_for_compiled_tracer_counter)
    start_time = perf_counter()
    while trc.iter():
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('PASS - Compiled Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    trc = TracerCounterCompiled(tr.iter_per_time_unit, test_time, clock_type_for_compiled_tracer_counter)
    start_time = perf_counter()
    for i in TracerIteratorCompiled(trc):
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('PASS - Compiled Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    print()
else:
    clock_type_for_native = ClockTypeNative.perf_counter
    clock_type_for_native_tracer = clock_type_for_native
    clock_type_for_native_tracer_counter = ClockTypeNative.fake

    tr = TracerNative(test_time, clock_type_for_native_tracer)
    start_time = perf_counter()
    while tr.iter():
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = tr.iterations_made / time_spent
    print('Native .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second,
                                                                                                    time_spent,
                                                                                                    tr.iter_per_time_unit,
                                                                                                    tr.time_spent,
                                                                                                    tr.iterations_made))

    trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
    start_time = perf_counter()
    while trc.iter():
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('Native Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(iter_per_second,
                                                                                                    time_spent,
                                                                                                    trc.iter_per_time_unit,
                                                                                                    trc.time_spent,
                                                                                                    trc.iterations_made))

    trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
    start_time = perf_counter()
    for i in TracerIteratorNative(trc):
        k = int('1243')
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('Native Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(
        iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    tr = TracerNative(test_time, clock_type_for_native_tracer)
    start_time = perf_counter()
    while tr.iter():
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = tr.iterations_made / time_spent
    print('PASS - Native .iter(): {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(
        iter_per_second, time_spent, tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
    start_time = perf_counter()
    while trc.iter():
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('PASS - Native Counter: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(
        iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    trc = TracerCounterNative(tr.iter_per_time_unit, test_time, clock_type_for_native_tracer_counter)
    start_time = perf_counter()
    for i in TracerIteratorNative(trc):
        pass
    end_time = perf_counter()
    time_spent = end_time - start_time
    iter_per_second = trc.iterations_made / time_spent
    print('PASS - Native Counter Iterator: {} iter/s; {} seconds. Internal: {} iter/s; {} seconds; {} iters'.format(
        iter_per_second, time_spent, trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    print()

# =============================================================

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
for i in range(iter_qnt):
    k = int('1243')
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt / time_spent
print('for i in range(): {} iter/s'.format(iter_per_second))

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
for i in range(iter_qnt):
    pass
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt / time_spent
print('PASS - for i in range(): {} iter/s'.format(iter_per_second))

print()

# =============================================================

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
iter_qnt_buff = iter_qnt
while iter_qnt > 0:
    k = int('1243')
    iter_qnt -= 1
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt_buff / time_spent
print('while iter_qnt > 0: {} iter/s'.format(iter_per_second))

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
iter_qnt_buff = iter_qnt
while iter_qnt > 0:
    pass
    iter_qnt -= 1
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt_buff / time_spent
print('PASS - while iter_qnt > 0: {} iter/s'.format(iter_per_second))

print()

# =============================================================

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
iter_qnt_buff = iter_qnt
while True:
    k = int('1243')
    iter_qnt -= 1
    if iter_qnt <= 0:
        break
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt_buff / time_spent
print('while True: {} iter/s'.format(iter_per_second))

start_time = perf_counter()
iter_qnt = round(tr.iter_per_time_unit)
iter_qnt_buff = iter_qnt
while True:
    pass
    iter_qnt -= 1
    if iter_qnt <= 0:
        break
end_time = perf_counter()
time_spent = end_time - start_time
iter_per_second = iter_qnt_buff / time_spent
print('PASS - while True: {} iter/s'.format(iter_per_second))

print()

# =============================================================



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__author__ = 'Mikhail Butenko <gtalk@mikhail-butenko.in.ua>'

from cengal.data_generation.id_generator import IDGenerator
# from help_tools import AddToCompoundDict, ResultExistence
from cengal.code_flow_control.smart_values.versions.v_0 import ResultExistence
from cengal.data_containers.compound_dict_management import AddToCompoundDict


SMART_TREE_TYPE = 0  # smart tree. Умное дерево тегов: сеть отображенная на древо. Возвращает только список
    # непосредственных подтегов текущего пути, но не их подтеги; возвращает элементы текущего пути, но не элементы
    # из подпутей
SMART_TREE_TYPE_WITH_INTERNAL_MENU = 1   # smart tree with internal menu. В древо встроено меню, позволяющее прямо из
    # древа производить смену типа вывода: SMART_TREE_TYPE, FULL_TREE_TYPE и PLAIN_PSEUDO_TREE_TYPE. На каждый тип
    # вывода будет доступен подтег/подпапка, внутри когорого уже будет нормальное древо элементов, но уже выбранного
    # типа
FULL_TREE_TYPE = 2  # full tree with all tags - with repeats and without filtering. Список айтемов - как у
    # SMART_TREE_TYPE, но при этом список тегов - как у PLAIN_PSEUDO_TREE_TYPE
PLAIN_PSEUDO_TREE_TYPE = 3   # plain tags and items set (will show all tags, subtags and items of
    # current hm... dir - current tag set). Показывает все теги и подтеги единым списком - как у примитивных теговых
    # файловых систем; показывает все элементы текущего пути + все элементы всех под-путей

USUAL_TREE_TYPE = PLAIN_PSEUDO_TREE_TYPE

_ROOT_TAG = r'k{1+vdcY#m8t-4m9`)G2\b]/O\'Rzqyr@FEO~%./nGPzl)[^q 0RS!.bCPh ?fag{8~{SGj;Ss3U85Q-:'


class ToManyIdenticalItemsOnTheGivenTagPathError(Exception):
    pass


class UnknownTreeTypeError(Exception):
    pass


class AddToCompoundDict__Set(AddToCompoundDict):
    def __init__(self, original_dict):
        super().__init__(
            original_dict,
            lambda: set(),
            lambda original_dict, key, value: original_dict[key].add(value)
        )


class TagDB:

    def __init__(self):
        super().__init__()

        self.itemsID = IDGenerator()

        self.itemIDsForItem = {}  # key - item hash; data - set of itemIDs

        self.itemsSet = {}  # key - ItemID; data - binItem
        # TODO: заменить список тегов на хеш единожды сохраненного списка тегов
        self.itemWithTags = {}  # key - ItemID; data - sorted common TagsTuple's hash

        # TODO: убрать tagsNumPerItemID из кода. Заменить этот список itemID - на список hashOfTheTagHashTuple
        self.tagsNumPerItemID = {}  # key - number of tags in this ItemID group; data - set of itemIDs which are have
            # needed number of tags

        self.tagsSet = {}  # key - tag hash; data - binTag
        self.tagWithItems = {}  # key - Tag hash; data - set of itemIDs
        self.tagsQnt = {}    # key - Tag hash; data - quantity of the items with this tag


        self.commonTagSets = {}  # key - sorted common TagsTuple's hash; data - sorted TagsTuple
        self.itemsOnTheCommonTagSets = {}   # key - sorted common TagsTuple's hash; data - set of itemIDs
        self.tagsQntPerCommonTagSet = {}    # key - number of tags; data - set of TagsTuple hashes
        self.setOfTagGroupQnt = set()   # {tagQntInGroup1, tagQntInGroup2, ..., tagQntInGroupN} where
            # each Group is an key of the self.tagsQntPerCommonTagSet

        # TODO: добавить словарь tagHashTuplesIDs вида [tagHashTuple.__hash__():tagHashTuple]
        # где tagHashSet - это {tagHash_1, tagHash_2, ..., tagHash_3}
        # TODO: добавить словарь itemsOnThePath вида [tagHashTuple.__hash__():{itemID_1, itemID_2, ..., itemID_3}]

        # TODO: добавить словарь вида [tagQnt:{itemID_1, itemID_2, ..., itemID_3}] и вычитывать это из него
        # и/или
        # TODO: добавить словарь вида [tagQnt:{hashOfTheTagHashTuple_1, hashOfTheTagHashTuple_2, ...
        # , hashOfTheTagHashTuple_N}] и вычитывать это из него, а потом уже и из каждого tagHashSet
        # где hashOfTheTagHashTuple - это tagHashTuple.__hash__()

    def get_root_tag(self):
        return str(_ROOT_TAG)

    def add_tag(self, binTag):
        tagHash = binTag.__hash__()
        self.tagsSet[tagHash] = binTag
        if tagHash not in self.tagsSet:
            self.tagsQnt[tagHash] = 0
        if tagHash not in self.tagWithItems:
            self.tagWithItems[tagHash] = set()

    def remove_tag(self, binTag):
        # will try to delete given tag. If there is at least one item with this tag, than function will fail
        # and will return False; otherwise it will delete given tag and will return True.
        functionResult = False
        tagHash = binTag.__hash__()
        if tagHash in self.tagsSet:
            if tagHash in self.tagsQnt:
                if self.tagsQnt[tagHash] == 0:
                    del self.tagsSet[tagHash]
                    del self.tagsQnt[tagHash]
                    functionResult = True
                else:
                    functionResult = False
            else:
                del self.tagsSet[tagHash]
                functionResult = True

        if functionResult:
            if tagHash in self.tagWithItems:
                del self.tagWithItems[tagHash]

        return functionResult

    def add_item(self, binItem, binTags):
        # will add new item and return it's dynamic ID or None object If this Item already exist on the given tag path
        # Or will raise an exception if we already have more than one binItem (another item that is identical to
        # the given binItem)  on this tag path

        binTags = set(binTags)

        if self.get_root_tag() not in binTags:
            binTags.add(self.get_root_tag())

        # may raise an exception in this place. Nope - from now it will be not
        if self.get_itemID_from_item_and_tags(binTags, binItem) is not None:
            return None

        itemID = self.itemsID()
        self.itemsSet[itemID] = binItem

        itemHash = binItem.__hash__()
        if itemHash in self.itemIDsForItem:
            IDsSet = self.itemIDsForItem[itemHash]
            IDsSet.add(itemID)
            # self.itemIDsForItem[itemHash] = IDsSet
        else:
            self.itemIDsForItem[itemHash] = {itemID}

        tagQnt = len(binTags)
        if tagQnt in self.tagsNumPerItemID:
            itemIDsSet = self.tagsNumPerItemID[tagQnt]
            itemIDsSet.add(itemID)
            # self.tagsNumPerItemID[tagQnt] = itemIDsSet
        else:
            self.tagsNumPerItemID[tagQnt] = {itemID}

        binTagHashes = set()

        for tag in binTags:
            self.add_tag(tag)
            tagHash = tag.__hash__()
            binTagHashes.add(tagHash)
            setOfItems = self.tagWithItems[tagHash]
            if itemID not in setOfItems:
                setOfItems.add(itemID)
                if tagHash in self.tagsQnt:
                    self.tagsQnt[tagHash] += 1
                else:
                    self.tagsQnt[tagHash] = 1
            self.tagWithItems[tagHash] = setOfItems

        sortedTagTuple = tuple(self.sort_tag_hash_list_by_hash(binTagHashes))
        hashOfTheSortedTagTuple = sortedTagTuple.__hash__()
        self.itemWithTags[itemID] = hashOfTheSortedTagTuple

        self.commonTagSets[hashOfTheSortedTagTuple] = sortedTagTuple

        if hashOfTheSortedTagTuple in self.itemsOnTheCommonTagSets:
            itemIDsSet = self.itemsOnTheCommonTagSets[hashOfTheSortedTagTuple]
            itemIDsSet.add(itemID)
            # self.itemsOnTheCommonTagSets[tagQnt] = itemIDsSet
        else:
            self.itemsOnTheCommonTagSets[hashOfTheSortedTagTuple] = {itemID}

        lenOfTheSortedTagTuple = len(sortedTagTuple)
        if lenOfTheSortedTagTuple in self.tagsQntPerCommonTagSet:
            itemIDsSet = self.tagsQntPerCommonTagSet[lenOfTheSortedTagTuple]
            itemIDsSet.add(hashOfTheSortedTagTuple)
            # self.tagsQntPerCommonTagSet[lenOfTheSortedTagTuple] = itemIDsSet
        else:
            self.tagsQntPerCommonTagSet[lenOfTheSortedTagTuple] = {hashOfTheSortedTagTuple}
            self.setOfTagGroupQnt.add(lenOfTheSortedTagTuple)

        return itemID

    def remove_item_by_itemID(self, itemID):
        if itemID in self.itemsSet:
            itemHash = self.itemsSet[itemID].__hash__()
            del self.itemsSet[itemID]

            if itemHash in self.itemIDsForItem:
                IDsSet = self.itemIDsForItem[itemHash]
                IDsSet.difference_update({itemID})
                # self.itemIDsForItem[itemHash] = IDsSet
                if len(IDsSet) < 1:
                    del self.itemIDsForItem[itemHash]

        if itemID in self.itemWithTags:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagTuple = self.commonTagSets[commonTagTupleHash]
            numberOfTags = len(tagTuple)

            if commonTagTupleHash in self.itemsOnTheCommonTagSets:
                IDsSet = self.itemsOnTheCommonTagSets[commonTagTupleHash]
                IDsSet.difference_update({itemID})
                # self.itemsOnTheCommonTagSets[commonTagTupleHash] = IDsSet
                if len(IDsSet) < 1:
                    del self.itemsOnTheCommonTagSets[commonTagTupleHash]
                    del self.commonTagSets[commonTagTupleHash]
                    if numberOfTags in self.tagsQntPerCommonTagSet:
                        setOfTagTuplesHashes = self.tagsQntPerCommonTagSet[numberOfTags]
                        setOfTagTuplesHashes.difference_update({numberOfTags})
                        # self.tagsQntPerCommonTagSet[numberOfTags] = setOfTagTuplesHashes
                        if len(setOfTagTuplesHashes) < 1:
                            del self.tagsQntPerCommonTagSet[numberOfTags]
                            self.setOfTagGroupQnt.difference_update({numberOfTags})

            del self.itemWithTags[itemID]

            setOfTagHashes = set(tagTuple)

            tagsQnt = len(setOfTagHashes)
            if tagsQnt in self.tagsNumPerItemID:
                IDsSet = self.tagsNumPerItemID[tagsQnt]
                IDsSet.difference_update({itemID})
                # self.tagsNumPerItemID[tagsQnt] = IDsSet
                if len(IDsSet) < 1:
                    del self.tagsNumPerItemID[tagsQnt]

            for tagHash in setOfTagHashes:
                if tagHash in self.tagsQnt:
                    tagsQuantity = self.tagsQnt[tagHash]
                    tagsQuantity -= 1
                    if tagsQuantity < 1:
                        tagsQuantity = 0
                    self.tagsQnt[tagHash] = tagsQuantity
                    if tagsQuantity < 1:
                        del self.tagsQnt[tagHash]
                        del self.tagsSet[tagHash]
                if tagHash in self.tagWithItems:
                    IDsSet = self.tagWithItems[tagHash]
                    IDsSet.difference_update({itemID})
                    if len(IDsSet) < 1:
                        del self.tagWithItems[tagHash]
                    # self.tagWithItems[tagHash] = IDsSet

        self.itemsID.remove_id(itemID)

    def remove_item(self, binTags, binItem):
        # will return ItemId for deleted item or None object if Item is not exist
        # Or will raise an exception if we already have more than one binItem (another item that is identical to
        # the given binItem) on this tag path
        binTags = set(binTags)
        if self.get_root_tag() not in binTags:
            binTags.add(self.get_root_tag())
        itemID = self.get_itemID_from_item_and_tags(binTags, binItem)
        if itemID is not None:
            self.remove_item_by_itemID(itemID)
        return itemID

    def __OLD__get_itemID_from_item_and_tags(self, binTags, binItem):
        if self.get_root_tag() not in binTags:
            binTags.append(self.get_root_tag())
        potentialIDs = set(self.get_potential_itemIDs_from_item(binItem))
        itemIDsSet = set(self.get_itemIDs_from_tags(binTags, SMART_TREE_TYPE))
        resultItemIDsList = potentialIDs & itemIDsSet
        if len(resultItemIDsList) == 0:
            return None
        elif len(resultItemIDsList) == 1:
            resultItemID = resultItemIDsList.pop()  # we have assume that we'll have only one item in intersection
                # between potential IDs and Items that have (and have only) given tag list (without another tags in the
                # path to this items). We need to check it in the adding new item to the given tag path.
            return resultItemID
        elif len(resultItemIDsList) > 1:
            raise ToManyIdenticalItemsOnTheGivenTagPathError()

    def get_itemID_from_item_and_tags(self, binTags, binItem):
        binTags = set(binTags)

        if self.get_root_tag() not in binTags:
            binTags.add(self.get_root_tag())

        potentialIDs = self.get_potential_itemIDs_from_item(binItem)
        setOfBinTagsHashes = set()
        for tag in binTags:
            setOfBinTagsHashes.add(tag.__hash__())
        for itemID in potentialIDs:
            currentItemTagsSet = self.get_tagsHashes_from_single_item(itemID, isWithoutRootHash=False)
            if setOfBinTagsHashes == currentItemTagsSet:
                return itemID
        return None

    def tag_hash_list_2_tag_list(self, tagHashList):
        tagList = list()
        for tagHash in tagHashList:
            tagList.append(self.tagsSet[tagHash])
        return tagList

    def get_item_and_tags_from_itemID(self, itemID):
        commonTagTupleHash = self.itemWithTags[itemID]
        tagSet = set(self.commonTagSets[commonTagTupleHash]) - {self.get_root_tag().__hash__()}
        sortedTagHashList = self.sort_tag_hash_list_by_qnt(tagSet)
        result = (self.itemsSet[itemID], self.tag_hash_list_2_tag_list(sortedTagHashList))
        return result

    # @profile
    def get_top_tag_hash_list_by_qnt(self, tagHashSet, local_tags_qnt=None):
        tagsQnt = self.tagsQnt
        if local_tags_qnt is not None:
            tagsQnt = local_tags_qnt
        tag_hash_set = set(tagHashSet)
        tag_by_qnt = dict()
        tag_by_qnt__filler = AddToCompoundDict__Set(tag_by_qnt)
        biggest_qnt = 0
        for tag_hash in tag_hash_set:
            qnt = tagsQnt[tag_hash]
            if qnt > biggest_qnt:
                biggest_qnt = qnt
            # if qnt not in tag_by_qnt:
            #     tag_by_qnt[qnt] = set()
            # tag_by_qnt[qnt].add(tag_hash)
            tag_by_qnt__filler.add(qnt, tag_hash)

        result = None
        if len(tag_by_qnt) > 0:
            # biggest_qnt = max(tag_by_qnt)
            result = tuple(tag_by_qnt[biggest_qnt])
        else:
            result = tuple()
        return result

    def sort_tag_hash_list_by_qnt(self, tagHashSet):
        # will return sorted tag list - not sorted tag hash list
        tagHashSet = set(tagHashSet)
        rawTagList = list()
        for tagHash in tagHashSet:
            tagWithWeight = (tagHash, self.tagsQnt[tagHash])
            rawTagList.append(tagWithWeight)
        return self.sort_raw_tag_list(rawTagList)

    def sort_tag_list_by_qnt(self, binTags):
        binTags = set(binTags)
        rawTagList = list()
        for tag in binTags:
            tagHash = tag.__hash__()
            tagWithWeight = (tag, self.tagsQnt[tagHash])
            rawTagList.append(tagWithWeight)
        return self.sort_raw_tag_list(rawTagList)

    def sort_tag_hash_list_by_hash(self, tagHashSet):
        # will return sorted tag list - not sorted tag hash list
        tagHashSet = set(tagHashSet)
        rawTagList = list()
        for tagHash in tagHashSet:
            tagWithWeight = (tagHash, tagHash)
            rawTagList.append(tagWithWeight)
        return self.sort_raw_tag_list(rawTagList)

    def sort_tag_list_by_hash(self, binTags):
        binTags = set(binTags)
        rawTagList = list()
        for tag in binTags:
            tagHash = tag.__hash__()
            tagWithWeight = (tag, tagHash)
            rawTagList.append(tagWithWeight)
        return self.sort_raw_tag_list(rawTagList)

    def sort_raw_tag_list(self, rawTagList):
        # will return sorted tag list
        rawTagList = sorted(rawTagList, key=lambda tagAndWeight: tagAndWeight[1], reverse=True)
        tagList = list()
        for rawTag in rawTagList:
            tagList.append(rawTag[0])
        return tagList

    def get_itemIDs_from_tags(self, binTags, treeType=USUAL_TREE_TYPE,
                              isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags=False):
        # TODO: исправить ошибку: SMART_TREE_TYPE: возвращает не только список файлов в текущей директории, но и из
        # непосредственных подпапок данной папки

        # treeType - type of the graph tree representation: show all tags with replies (pure representation);
        # show only relevant tags; etc.
        # return set of itemIDs

        binTags = set(binTags)

        if self.get_root_tag() not in binTags:
            binTags.add(self.get_root_tag())

        tagHashSet = set()
        for binTag in binTags:
            tagHashSet.add(binTag.__hash__())

        # PLAIN_PSEUDO_TREE_TYPE
        interceptionOfItemsWithTags = set()
        if (treeType == PLAIN_PSEUDO_TREE_TYPE) or isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags:
            binTagsQnt = len(tagHashSet)
            commonTagGroupHashSet = set()
            tagSet = set()
            itemIDSet = set()
            binTagHashTuple = tuple(self.sort_tag_hash_list_by_hash(tagHashSet))
            hashOfTheBinTagHashTuple = binTagHashTuple.__hash__()
            if hashOfTheBinTagHashTuple in self.itemsOnTheCommonTagSets:
                itemIDSet = self.itemsOnTheCommonTagSets[hashOfTheBinTagHashTuple]
            for commonTagQnt in self.setOfTagGroupQnt:
                if commonTagQnt > binTagsQnt:
                    setOfTheCommonTagGroupHashes = self.tagsQntPerCommonTagSet[commonTagQnt]
                    commonTagGroupHashSet.update(setOfTheCommonTagGroupHashes)
            for commonTagGroupHash in commonTagGroupHashSet:
                commonTagHashTuple = self.commonTagSets[commonTagGroupHash]
                commonTagHashSet = set(commonTagHashTuple)
                if tagHashSet.issubset(commonTagHashSet):
                    itemIDSet.update(self.itemsOnTheCommonTagSets[commonTagGroupHash])
                # # if len(tagHashSet & commonTagHashSet) == len(tagHashSet):
                # res_set = tagHashSet.intersection(commonTagHashSet)
                # if len(res_set) == binTagsQnt:
                #     itemIDSet = itemIDSet | self.itemsOnTheCommonTagSets[commonTagGroupHash]
            interceptionOfItemsWithTags = itemIDSet

            # isFirstHash = True
            # for tag in binTags:
            #     tagHash = tag.__hash__()
            #     if tagHash in self.tagWithItems:
            #         if isFirstHash:
            #             interceptionOfItemsWithTags = self.tagWithItems[tagHash]
            #             isFirstHash = False
            #         else:
            #             itemsWithTag = self.tagWithItems[tagHash]
            #             interceptionOfItemsWithTags = interceptionOfItemsWithTags & itemsWithTag
            #     else:
            #         # TODO: произвести такую же провеку в get_items_from_tags() и build_smart_tree()
            #         if isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags:
            #             result = (set(), set())
            #             return result
            #         else:
            #             return set()

        resultItemIDSet = interceptionOfItemsWithTags
        setOfAllInternalItemIDsForThisSetOfTags = interceptionOfItemsWithTags

        # SMART_TREE_TYPE or FULL_TREE_TYPE
        if (treeType == SMART_TREE_TYPE) or (treeType == FULL_TREE_TYPE):
            resultItemIDSet = set()
            binTagHashTuple = self.sort_tag_hash_list_by_hash(tagHashSet)
            binTagHashTuple = tuple(binTagHashTuple)
            hashOfTheBinTagHashTuple = binTagHashTuple.__hash__()
            if hashOfTheBinTagHashTuple in self.itemsOnTheCommonTagSets:
                resultItemIDSet = self.itemsOnTheCommonTagSets[hashOfTheBinTagHashTuple]

            # filteredItemIDsSet = set()
            # tagQnt = len(binTags)
            # # for itemID in setOfAllInternalItemIDsForThisSetOfTags:
            # #     if len(self.itemWithTags[itemID]) == tagQnt:
            # #         # _TODO: добавить словарь вида [tagQnt:{itemID_1, itemID_2, ..., itemID_3}]
            # #         # и вычитывать это из него
            # #         # и/или
            # #         # _TODO: добавить словарь вида [tagQnt:{hashOfTheTagHashSet_1, hashOfTheTagHashSet_2, ...
            # #         # , hashOfTheTagHashSet_3}] и вычитывать это из него, а потом уже и из каждого tagHashSet
            # #         # где hashOfTheTagHashSet - это tagHashSet.__hash__()
            # #         filteredItemIDsSet.add(itemID)
            # if tagQnt in self.tagsNumPerItemID:
            #     filteredItemIDsSet = setOfAllInternalItemIDsForThisSetOfTags & self.tagsNumPerItemID[tagQnt]
            #
            # resultItemIDSet = set()
            # tagHashSet = set()
            # for binTag in binTags:
            #     tagHashSet.add(binTag.__hash__())
            # for itemID in filteredItemIDsSet:
            #     commonTagTupleHash = self.itemWithTags[itemID]
            #     tagSet = set(self.commonTagSets[commonTagTupleHash])
            #     if tagSet == tagHashSet:
            #         # _TODO: добавить словарь tagHashTuplesIDs вида [tagHashTuple.__hash__():tagHashTuple]
            #         # где tagHashSet - это {tagHash_1, tagHash_2, ..., tagHash_3}
            #         # _TODO: добавить словарь itemsOnThePath вида [tagHashTuple.__hash__():{itemID_1, itemID_2, ...
            #         # , itemID_3}]
            #         resultItemIDSet.add(itemID)
        elif treeType == PLAIN_PSEUDO_TREE_TYPE:
            # already implemented (see bellow). Don't touch this code!
            pass
        else:
            raise UnknownTreeTypeError()

        if isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags:
            result = (set(resultItemIDSet), set(setOfAllInternalItemIDsForThisSetOfTags))
            return result
        else:
            return set(resultItemIDSet)

    def get_items_from_tags(self, binTags, treeType=USUAL_TREE_TYPE,
                            isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags=False):
        # treeType - type of the graph tree representation: show all tags with replies (pure representation);
        # show only relevant tags; etc.
        # return set of itemIDs
        binTags = set(binTags)
        itemIDsSet = self.get_itemIDs_from_tags(binTags, treeType=treeType,
                                                isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags=
                                                isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags)
        if isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags:
            itemSet = set()
            for itemID in itemIDsSet[0]:
                itemSet.add(self.itemsSet[itemID])
            result = (tuple(itemSet), tuple(itemIDsSet[1]))  # result == (usual items set, additional set of all
                # internal itemIDs)
            return result
        else:
            itemSet = set()
            for itemID in itemIDsSet:
                itemSet.add(self.itemsSet[itemID])
            return tuple(itemSet)

    def get_tagHashes_from_tags(self, binTags, treeType=USUAL_TREE_TYPE,
                                prePreparedSetOfAllInternalItemIDsForThisSetOfTags=None):
        # where "itemIDsSet" is externally given "get_itemIDs_from_tags(binTags, treeType=PLAIN_PSEUDO_TREE_TYPE)"
        # so "itemIDsSet" is a set of the all items inside the "folder" binTags (including items from "subfolders")
        # treeType - the same as in the "get_items_from_tags()" method
        # return set of itemIDs
        # prePreparedSetOfAllInternalItemIDsForThisSetOfTags can be generated by:
        #   a) get_itemIDs_from_tags(..., treeType=PLAIN_PSEUDO_TREE_TYPE)
        #   a) get_itemIDs_from_tags(..., isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags = True)
        #   c) get_items_from_tags(..., isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags = True)
        #   d) get_items_from_tags(..., treeType=PLAIN_PSEUDO_TREE_TYPE) || BUT: it'll return item set - not itemID
        #       set

        binTags = set(binTags)

        resultTagHashSet = set()

        setOfAllInternalItemIDs = set()
        if prePreparedSetOfAllInternalItemIDsForThisSetOfTags is None:
            setOfAllInternalItemIDs = self.get_itemIDs_from_tags(binTags, treeType=PLAIN_PSEUDO_TREE_TYPE)
        else:
            setOfAllInternalItemIDs = prePreparedSetOfAllInternalItemIDsForThisSetOfTags

        if (treeType == PLAIN_PSEUDO_TREE_TYPE) or (treeType == FULL_TREE_TYPE):
            binTagHashes = set()
            for tag in binTags:
                binTagHashes.add(tag.__hash__())
            tagHashSet = set()
            for itemID in setOfAllInternalItemIDs:
                if itemID in self.itemWithTags:
                    commonTagTupleHash = self.itemWithTags[itemID]
                    tagSet = set(self.commonTagSets[commonTagTupleHash])
                    tagHashSet.update(tagSet)
            resultTagHashSet = tagHashSet - binTagHashes
        elif treeType == SMART_TREE_TYPE:
            # smartTree = self.build_smart_tree(binTags, prePreparedSetOfAllInternalItemIDs=setOfAllInternalItemIDs)
            smartTree = self.build_smart_tree(binTags, prePreparedSetOfAllInternalItemIDs=setOfAllInternalItemIDs,
                                              zeroSliceOnly=True)
            if 0 in smartTree:
                resultTagHashSet = smartTree[0]
            # filteredItemIDsList = list()
            # tagQnt = len(binTags)
            # for itemID in listOfAllInternalItemIDs:
            #     if len(self.itemWithTags[itemID]) == (tagQnt + 1):
            #         filteredItemIDsList.append(itemID)
            #
            # tagHashSet = set()
            # for itemID in filteredItemIDsList:
            #     tagHashSet.update(set(self.itemWithTags[itemID]))
            # givenTagHashes = set()
            # for tag in binTags:
            #     givenTagHashes.add(tag.__hash__())
            # tagHashSet.difference_update(givenTagHashes)
            # ##resultTagHashList = list(tagHashSet)
            # # если остановиться тут - то мы увидим не все папки: мы не увидим папки непосредственно в которых есть
            # # только другие подпапки, но ни одного файла.
            # #
            # # значит далее мы должны исключить все файлы, которые имеют только что найденные теги, и начать строить
            # # древо тегов для оставшихся
            # #
            # # а далее - повторить все это в цикле, увеличив при проверке кол-во тегов еще раз на единицу (и используя
            # # уже оставшийся после отсеивания набор файлов). В итоге кол-во итераций зависит не от количества файлов,
            # # а от максимальной фактически имеющейся вложенности файлов внутри тегов-каталогов
        else:
            raise UnknownTreeTypeError()

        sortedTagHashList = self.sort_tag_hash_list_by_qnt(resultTagHashSet - {self.get_root_tag().__hash__()})
        return list(sortedTagHashList)

    def get_tags_from_tags(self, binTags, treeType=USUAL_TREE_TYPE,
                           prePreparedSetOfAllInternalItemIDsForThisSetOfTags=None):
        result = self.get_tagHashes_from_tags(binTags, treeType=treeType,
                                              prePreparedSetOfAllInternalItemIDsForThisSetOfTags=
                                              prePreparedSetOfAllInternalItemIDsForThisSetOfTags)
        return tuple(self.tag_hash_list_2_tag_list(result))

    def build_smart_tree(self, startingBinTags, prePreparedSetOfAllInternalItemIDs=None, zeroSliceOnly=False):
        startingBinTags = set(startingBinTags)

        if self.get_root_tag() not in startingBinTags:
            startingBinTags.add(self.get_root_tag())

        startingTagHashes = set()
        for tag in startingBinTags:
            startingTagHashes.add(tag.__hash__())

        if prePreparedSetOfAllInternalItemIDs is None:
            setOfAllInternalItemIDs = self.get_itemIDs_from_tags(startingBinTags, treeType=PLAIN_PSEUDO_TREE_TYPE)
        else:
            setOfAllInternalItemIDs = prePreparedSetOfAllInternalItemIDs

        smartTree = {0: set()}
        smartTree__filler = AddToCompoundDict__Set(smartTree)
        local_tags_qnt = dict()
        local_tags_qnt__filler = AddToCompoundDict(
            local_tags_qnt,
            lambda: 0,
            lambda container, key, value: (True, container[key] + 1)
        )
        for itemID in setOfAllInternalItemIDs:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagSet = set(self.commonTagSets[commonTagTupleHash])
            setOfTags = tagSet
            setOfTags = setOfTags - startingTagHashes
            for tag_hash in setOfTags:
                # if tag_hash not in local_tags_qnt:
                #     local_tags_qnt[tag_hash] = 0
                # local_tags_qnt[tag_hash] += 1
                local_tags_qnt__filler.add(tag_hash)

        for itemID in setOfAllInternalItemIDs:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagSet = set(self.commonTagSets[commonTagTupleHash])
            setOfTags = tagSet
            setOfTags = setOfTags - startingTagHashes
            listOfTagHashes = None
            if zeroSliceOnly:
                listOfTagHashes = self.get_top_tag_hash_list_by_qnt(setOfTags, local_tags_qnt)
            else:
                listOfTagHashes = self.sort_tag_hash_list_by_qnt(setOfTags)

            lastTagHash = None
            lastTagHashQnt = None
            treeLevel = 0
            for tagHash in listOfTagHashes:
                # currentTagHashQnt = self.tagsQnt[tagHash]
                currentTagHashQnt = local_tags_qnt[tagHash]
                if (lastTagHash is None) or (currentTagHashQnt == lastTagHashQnt):
                    pass
                else:
                    treeLevel += 1
                    lastTagHash = None
                    lastTagHashQnt = None
                # if treeLevel not in smartTree:
                #     smartTree[treeLevel] = set()
                # # tagsSetOnTheLevel = smartTree[treeLevel]
                # # tagsSetOnTheLevel.add(tagHash)
                # # smartTree[treeLevel] = tagsSetOnTheLevel
                # smartTree[treeLevel].add(tagHash)
                smartTree__filler.add(treeLevel, tagHash)
                lastTagHash = tagHash
                lastTagHashQnt = currentTagHashQnt
        return dict(smartTree)

    def build_smart_tree_2(self, startingBinTags, prePreparedSetOfAllInternalItemIDs=None, zeroSliceOnly=False):
        startingBinTags = set(startingBinTags)

        if self.get_root_tag() not in startingBinTags:
            startingBinTags.add(self.get_root_tag())

        startingTagHashes = set()
        for tag in startingBinTags:
            startingTagHashes.add(tag.__hash__())

        if prePreparedSetOfAllInternalItemIDs is None:
            setOfAllInternalItemIDs = self.get_itemIDs_from_tags(startingBinTags, treeType=PLAIN_PSEUDO_TREE_TYPE)
        else:
            setOfAllInternalItemIDs = prePreparedSetOfAllInternalItemIDs

        smartTree = {0: set()}
        smartTree__filler = AddToCompoundDict__Set(smartTree)
        local_tags_qnt = dict()
        local_tags_qnt__filler = AddToCompoundDict(
            local_tags_qnt,
            lambda: 0,
            lambda container, key, value: (True, container[key] + 1)
        )
        for itemID in setOfAllInternalItemIDs:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagSet = set(self.commonTagSets[commonTagTupleHash])
            setOfTags = tagSet
            setOfTags = setOfTags - startingTagHashes
            for tag_hash in setOfTags:
                # if tag_hash not in local_tags_qnt:
                #     local_tags_qnt[tag_hash] = 0
                # local_tags_qnt[tag_hash] += 1
                local_tags_qnt__filler.add(tag_hash)

        for itemID in setOfAllInternalItemIDs:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagSet = set(self.commonTagSets[commonTagTupleHash])
            setOfTags = tagSet
            setOfTags = setOfTags - startingTagHashes
            listOfTagHashes = None
            if zeroSliceOnly:
                listOfTagHashes = self.get_top_tag_hash_list_by_qnt(setOfTags, local_tags_qnt)
            else:
                listOfTagHashes = self.sort_tag_hash_list_by_qnt(setOfTags)

            lastTagHash = None
            lastTagHashQnt = None
            treeLevel = 0
            for tagHash in listOfTagHashes:
                # currentTagHashQnt = self.tagsQnt[tagHash]
                currentTagHashQnt = local_tags_qnt[tagHash]
                if (lastTagHash is None) or (currentTagHashQnt == lastTagHashQnt):
                    pass
                else:
                    treeLevel += 1
                    lastTagHash = None
                    lastTagHashQnt = None
                # if treeLevel not in smartTree:
                #     smartTree[treeLevel] = set()
                # # tagsSetOnTheLevel = smartTree[treeLevel]
                # # tagsSetOnTheLevel.add(tagHash)
                # # smartTree[treeLevel] = tagsSetOnTheLevel
                # smartTree[treeLevel].add(tagHash)
                smartTree__filler.add(treeLevel, tagHash)
                lastTagHash = tagHash
                lastTagHashQnt = currentTagHashQnt
        return dict(smartTree)

    def get_all_from_tags(self, binTags, treeType=USUAL_TREE_TYPE):
        binTags = set(binTags)
        items = self.get_items_from_tags(binTags, treeType=treeType,
                                         isAlsoNeedSetOfAllInternalItemIDsForThisSetOfTags=True)
        if len(items[1]) > 0:
            tags = self.get_tags_from_tags(binTags, treeType=treeType,
                                           prePreparedSetOfAllInternalItemIDsForThisSetOfTags=items[1])
            result = (tuple(tags), set(items[0]))
            return result
        else:
            result = (tuple(), set())
            return result

    def get_tagsHashes_from_single_item(self, itemID, isWithoutRootHash=True):
        if itemID in self.itemWithTags:
            commonTagTupleHash = self.itemWithTags[itemID]
            tagSet = set(self.commonTagSets[commonTagTupleHash])
            if isWithoutRootHash:
                return set(tagSet - {self.get_root_tag().__hash__()})
            else:
                return set(tagSet)
        else:
            return set()

    def get_potential_itemIDs_from_item(self, binItem):
        itemHash = binItem.__hash__()
        if itemHash in self.itemIDsForItem:
            return set(self.itemIDsForItem[itemHash])
        else:
            return set()

    def is_smart_redirection_for_a_tag_path_reduction_needed(self, binTags):
        binTags = set(binTags)

        if self.get_root_tag() not in binTags:
            binTags.add(self.get_root_tag())

        tagHashSet = set()
        for binTag in binTags:
            tagHashSet.add(binTag.__hash__())

        binTagsQnt = len(tagHashSet)
        commonTagGroupHashSet = set()
        # setOfLenOfTheCommonTagHashSetForChecking = set()
        setOfTheTagsIntersection = None
        for commonTagQnt in self.setOfTagGroupQnt:
            if commonTagQnt > binTagsQnt:
                setOfTheCommonTagGroupHashes = self.tagsQntPerCommonTagSet[commonTagQnt]
                commonTagGroupHashSet.update(setOfTheCommonTagGroupHashes)
        for commonTagGroupHash in commonTagGroupHashSet:
            commonTagHashTuple = self.commonTagSets[commonTagGroupHash]
            commonTagHashSet = set(commonTagHashTuple)
            if len(tagHashSet & commonTagHashSet) == len(tagHashSet):
                if tagHashSet != commonTagHashSet:
                    if setOfTheTagsIntersection is None:
                        setOfTheTagsIntersection = commonTagHashSet
                    else:
                        setOfTheTagsIntersection = setOfTheTagsIntersection & commonTagHashSet
        #         if tagHashSet != commonTagHashSet:
        #             setOfLenOfTheCommonTagHashSetForChecking.add(len(commonTagHashSet))
        # minimalTagPath = min(setOfLenOfTheCommonTagHashSetForChecking)
        # pathDiff = minimalTagPath - len(tagHashSet)
        # if pathDiff > 0:
        if setOfTheTagsIntersection is None:
            return list()

        setOfTheTagsForAReduction = setOfTheTagsIntersection - tagHashSet

        sortedTagHashList = self.sort_tag_hash_list_by_qnt(setOfTheTagsForAReduction - {self.get_root_tag().__hash__()})
        return list(sortedTagHashList)

    def get_tags_for_a_smart_redirection(self, binTags):
        result = self.is_smart_redirection_for_a_tag_path_reduction_needed(binTags)
        return tuple(self.tag_hash_list_2_tag_list(result))



"""
compile_time_py_definitions.go

"""
# python wrapper for package cengal/_examples/ex_golang/versions/v_0/compilable/compile_time_py_definitions within overall package greeter
# This is what you import to use the package.
# File is generated by gopy. Do not edit.
# gopy build -output=compiled -vm=/usr/bin/python3 ./src ./compile_time_py_definitions

# the following is required to enable dlopen to open the _go.so file
import os,sys,inspect,collections
try:
	import collections.abc as _collections_abc
except ImportError:
	_collections_abc = collections

cwd = os.getcwd()
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
os.chdir(currentdir)
from . import _greeter
from . import go

os.chdir(cwd)

# to use this code in your end-user python file, import it as follows:
# from greeter import compile_time_py_definitions
# and then refer to everything using compile_time_py_definitions. prefix
# packages imported by this package listed below:




# ---- Types ---


#---- Enums from Go (collections of consts with same type) ---


#---- Constants from Go: Python can only ask that you please don't change these! ---
ApiEndpointDoc = "URL: https://api.production.com; Path: \\usr\\bin"
GOD_CENGAL_VERSION_MAJOR = 4
GOD_CENGAL_VERSION_MAJOR_STR = "4"
GOD_CENGAL_VERSION_MICRO = 0
GOD_CENGAL_VERSION_MICRO_STR = "0"
GOD_CENGAL_VERSION_MINOR = 4
GOD_CENGAL_VERSION_MINOR_STR = "4"
GOD_CENGAL_VERSION_STR = "4.4.0"
GOD_COMPILER_NAME = "x86_64-linux-gnu-gcc"
GOD_COMPILER_STRING = "x86_64-linux-gnu-gcc -pthread"
GOD_COMPILER_TYPE = "gcc"
GOD_CPU_ARCH = "X86_64"
GOD_CPU_ARCH_RAW_STRING = "x86_64"
GOD_CPU_BITS = 64
GOD_IS_ARM = 0
GOD_IS_ARM_BOOL = False
GOD_IS_BUILDING_FOR_PYODIDE = 0
GOD_IS_BUILDING_FOR_PYODIDE_BOOL = False
GOD_IS_INSIDE_OR_FOR_WEB_BROWSER = 0
GOD_IS_INSIDE_OR_FOR_WEB_BROWSER_BOOL = False
GOD_IS_RUNNING_IN_EMSCRIPTEN = 0
GOD_IS_RUNNING_IN_EMSCRIPTEN_BOOL = False
GOD_IS_RUNNING_IN_PYCHARM = 0
GOD_IS_RUNNING_IN_PYCHARM_BOOL = False
GOD_IS_RUNNING_IN_PYODIDE = 0
GOD_IS_RUNNING_IN_PYODIDE_BOOL = False
GOD_IS_RUNNING_IN_WASI = 0
GOD_IS_RUNNING_IN_WASI_BOOL = False
GOD_IS_X86 = 1
GOD_IS_X86_BOOL = True
GOD_KIVY_PLATFORM = "unknown"
GOD_KIVY_TARGET_PLATFORM = "unknown"
GOD_OS_API_TYPE = "posix"
GOD_OS_TYPE = "Linux"
GOD_PYTHON_IMPLEMENTATION = "CPython"
GOD_PYTHON_VERSION_MAJOR = 3
GOD_PYTHON_VERSION_MAJOR_STR = "3"
GOD_PYTHON_VERSION_MICRO = 10
GOD_PYTHON_VERSION_MICRO_STR = "10"
GOD_PYTHON_VERSION_MINOR = 8
GOD_PYTHON_VERSION_MINOR_STR = "8"
GOD_PYTHON_VERSION_STR = "3.8.10"
GOD_RAW_OS_PLATFORM = "linux"
GOD_UNAME_MACHINE = "x86_64"
GOD_UNAME_SYSNAME = "Linux"


# ---- Global Variables: can only use functions to access ---


# ---- Interfaces ---


# ---- Structs ---


# ---- Slices ---


# ---- Maps ---


# ---- Constructors ---


# ---- Functions ---





#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.data_manipulation.serialization import Serializer, Serializers, SerializerFeatures, DataFormats, Tags, best_serializer_for_standard_data
from enum import Enum
from typing import Dict, Any

from .commands import *


class SerializerID(Enum):
    multi_platform__initial_communication = -1
    multi_platform_fast = 0
    multi_platform = 1
    current_platform = 2
    current_platform__custom_types = 3


desired_features: Dict[SerializerID, SerializerFeatures] = {
    SerializerID.multi_platform__initial_communication.value: {
        DataFormats.json,
        Tags.deep,
        Tags.multi_platform,
    },
    SerializerID.multi_platform_fast.value: {
        DataFormats.any,
        Tags.deep,
        Tags.multi_platform,
        Tags.can_use_bytes,
    },
    SerializerID.multi_platform.value: {
        DataFormats.any,
        Tags.deep,
        Tags.multi_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
    },
    SerializerID.current_platform.value: {
        DataFormats.any,
        Tags.deep,
        Tags.can_use_set,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
    },
    SerializerID.current_platform__custom_types.value: {
        DataFormats.any,
        Tags.current_platform,
        Tags.deep,
        Tags.can_use_custom_types,
        Tags.can_use_bytes,
        Tags.can_use_set,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
    },
}


test_data: Dict[SerializerID, Any] = {
    SerializerID.multi_platform__initial_communication.value: {
        Fields.command_name.value: Commands.declare_client_node.value,
        Fields.request_id.value: 10,
        Fields.in_response_to.value: 10,
        Fields.is_response_required.value: True,
        Fields.data.value: {
            CommandDataFieldsDeclareClientNode.platform_name.value: 'Python 3.11.0',
            CommandDataFieldsDeclareClientNode.suggested_serializers.value: {
                SerializerID.multi_platform_fast.value: [
                    Serializers.msgpack_fast,
                    Serializers.orjson,
                    Serializers.cbor2,
                    Serializers.ujson,
                    Serializers.simplejson,
                    Serializers.json,
                ],
                SerializerID.multi_platform.value: [
                    Serializers.msgpack,
                    Serializers.orjson,
                    Serializers.cbor2,
                    Serializers.ujson,
                    Serializers.simplejson,
                    Serializers.json,
                ],
                SerializerID.current_platform.value: [
                    Serializers.marshal_compat_4,
                    Serializers.msgpack,
                    Serializers.orjson,
                    Serializers.cbor2,
                    Serializers.ujson,
                    Serializers.simplejson,
                    Serializers.json,
                    Serializers.cloudpickle_compat_5,
                    Serializers.cpickle_compat_5,
                    Serializers.pickle_compat_5,
                ],
                SerializerID.current_platform__custom_types.value: [
                    Serializers.cloudpickle_compat_5,
                    Serializers.cpickle_compat_5,
                    Serializers.pickle_compat_5,
                ],
            }
        }
    },
    SerializerID.multi_platform_fast.value: {
        DataFormats.any,
        Tags.deep,
        Tags.multi_platform,
        Tags.can_use_bytes,
    },
    SerializerID.multi_platform.value: {
        DataFormats.any,
        Tags.deep,
        Tags.multi_platform,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_list_as_list,
    },
    SerializerID.current_platform.value: {
        DataFormats.any,
        Tags.deep,
        Tags.can_use_set,
        Tags.can_use_bytes,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
    },
    SerializerID.current_platform__custom_types.value: {
        DataFormats.any,
        Tags.current_platform,
        Tags.deep,
        Tags.can_use_custom_types,
        Tags.can_use_bytes,
        Tags.can_use_set,
        Tags.decode_str_as_str,
        Tags.decode_bytes_as_bytes,
        Tags.decode_tuple_as_tuple,
        Tags.decode_list_as_list,
    },
}


def benchmark(test_data: Dict[SerializerID, Any]) -> Dict[SerializerID, Serializer]:
    ...



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform
from importlib import import_module
PYTHON_VERSION = platform.python_version_tuple()  # tuple() of str(); for example: ('3', '5', '1')
if PYTHON_VERSION[0] == '2':
    pass
elif PYTHON_VERSION[0] == '3':
    try:
        from imp import reload
    except ImportError:
        from importlib import reload
else:
    raise ImportError()

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


def reload_module(module):
    """
    You SHOULD use construction "import SOME_MODULE" before reload_module() call if construction
        "from SOME_MODULE import SOME_ITEM" was used originally.
    :param module:
    :return:
    """
    reload(module)


def reload_module_by_text_name(module_name, package=None):
    """
    :param module_name: Text representation of module name. "os.path" for example.
        See importlib.import_module() in the Python docs
    :param package: see importlib.import_module() in the Python docs
    :return:
    """
    module = import_module(module_name, package)
    reload(module)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.unittest.behavior_stabilizer import UnittestTestCaseWithState, UnittestTestCaseState
from cengal.performance_test_lib import PrecisePerformanceTestTracer
from cengal.introspection.inspect import func_name, pdi, pifrl
from threading import Lock, RLock
from pprint import pprint
from typing import Callable


def func_perf_test(func: Callable, *args, **kwargs):
    tr = PrecisePerformanceTestTracer(1.0)
    while tr.iter():
        func(*args, **kwargs)

    with tr as fast_iter:
        for i in fast_iter:
            func(*args, **kwargs)

    print('{}(): {} iter/s; {} seconds; {} iterations'.format(func_name(func), tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))


from cengal.hardware.memory.shared_memory import *
from cengal.time_management.cpu_clock import cpu_clock
from cengal.time_management.high_precision_sync_sleep import hps_sleep

import unittest


from dataclasses import dataclass
from typing import Any, Dict, Optional, Callable, Set, Type, Tuple, List, FrozenSet, Deque


class SimpleClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        
        return False


class SimpleClassWithSlots:
    __slots__ = ['a', 'b']

    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        
        return False


@dataclass
class DataClass:
    x: int
    y: 'Any'
    z: int = 5

    def add_one(self):
        self.x += 1


# Complex dataclass:

class Positions(IntEnum):
    manager = 0
    designer = 1
    accountant = 2
    lawyer = 3


class Employee:
    def __init__(self, name: str, age: int, position: Positions, years_of_employment: int = 0):
        self.name = name
        self.age = age
        self.position = position
        self.years_of_employment = years_of_employment
    
    def increase_years_of_employment(self):
        self.years_of_employment += 1
        self.age += 1
    
    def __repr__(self):
        return f'Employee(name={self.name}, age={self.age}, position={self.position}, years_of_employment={self.years_of_employment})'
    
    def __str__(self):
        return self.__repr__()


@dataclass
class CompanyInfo:
    company_id: int
    emails: Tuple[str, str]
    websites: List[str]
    income: float
    employees: int
    some_employee: Employee


@dataclass
class SomeSharedObject:
    some_processing_stage_control: bool
    str_value: str
    data_dict: Dict[Hashable, Any]
    company_info: CompanyInfo


# class TestCaseForRunInLoop(UnittestTestCaseWithState):
class TestCaseForRunInLoop(unittest.TestCase):
    repeats_num: int = 1

    # @classmethod
    # def setUpClass(cls):
    #     cls.state = UnittestTestCaseState()

    # @classmethod
    # def tearDownClass(cls):
    #     cls.state.unregister()
    
    # @unittest.skip('temp skip')
    def test_set(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: Set = {
                    'hello',
                    'world',
                }
                mapped_obj = creator.put_message(val)
                obj: Set = creator.get_message()
                obj |= set(range(20))
                obj.discard('hello')
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertIn('world', val)
                self.assertIn('world', obj)
                self.assertIn('world', mapped_obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('temp skip')
    def test_fast_set(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: Set = FastLimitedSet({
                    'hello',
                    'world',
                })
                mapped_obj = creator.put_message(val)
                obj: Set = creator.get_message()
                obj |= set(range(20))
                obj.discard('hello')
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertIn('world', val)
                self.assertIn('world', obj)
                self.assertIn('world', mapped_obj)

                self.assertNotEqual(val, obj)
                self.assertNotEqual(mapped_obj, obj)
                self.assertEqual(mapped_obj, val)
    
    # @unittest.skip('temp skip')
    def test_dict(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = {
                    'hello': 'world',
                    'world': 'hello',
                }
                mapped_obj = creator.put_message(val)
                obj: Dict = creator.get_message()
                obj.update({i: str(i) for i in range(20)})
                del obj['world']
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertIn('hello', val)
                self.assertIn('hello', obj)
                self.assertIn('hello', mapped_obj)

                self.assertIn('world', val)
                self.assertNotIn('world', obj)
                self.assertNotIn('world', mapped_obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('temp skip')
    def test_fast_dict(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = FastLimitedDict({
                    'hello': 'world',
                    'world': 'hello',
                })
                mapped_obj = creator.put_message(val)
                obj: Dict = creator.get_message()
                obj.update({i: str(i) for i in range(20)})
                obj['hello'] = 20
                del obj['world']
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertIn('hello', val)
                self.assertIn('hello', obj)
                self.assertIn('hello', mapped_obj)

                self.assertIn('world', val)
                self.assertNotIn('world', obj)
                self.assertIn('world', mapped_obj)

                self.assertNotEqual(val, obj)
                self.assertNotEqual(mapped_obj, obj)
                self.assertEqual(mapped_obj, val)
    
    # @unittest.skip('temp skip')
    def test_np_ndarray(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                shape = (10, 10, 3)
                dtype = np.float64
                val = np.zeros(shape, dtype=dtype)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()

                obj[0, 0, 0] = 100.0
                obj[1, 1, 1] = 200.0
                obj[2, 2, 2] = 300.0
                obj[3, 3, 0] = 400.0
                obj[4, 4, 1] = 500.0
                obj[5, 5, 2] = 600.0
                obj[6, 6, 0] = 700.0
                obj[7, 7, 1] = 800.0
                obj[8, 8, 2] = 900.0
                obj[9, 9, 0] = 1000.0

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                
                np.testing.assert_array_equal(mapped_obj, obj)
                try:
                    np.testing.assert_array_equal(val, mapped_obj)
                    assert False
                except AssertionError:
                    pass  # This is expected as we want them to be not equal
    
    # @unittest.skip('temp skip')
    def test_simple_class_inplace_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIs(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_simple_class_obj_default(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIs(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_simple_class_obj_static(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(ForceStaticObjectCopy(val))
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_simple_class_obj_general(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClass(1, 2)
                mapped_obj = creator.put_message(ForceGeneralObjectCopy(val))
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_decimal_obj(self):
        pifrl()
        from decimal import Decimal, DecimalTuple
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = [Decimal('3.14')]
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj[0] += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val[0], mapped_obj[0])
                self.assertIsNot(val[0], obj[0])

                self.assertNotEqual(val[0], obj[0])
                self.assertEqual(mapped_obj[0], obj[0])
    
    # @unittest.skip('broken')
    def test_simple_class_with_slots_obj_default(self):
        pifrl()
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = SimpleClassWithSlots(1, 2)
                mapped_obj = creator.put_message(val)
                obj = creator.get_message()
                obj.a += 1
                # print(f'{val.a=}, {obj.a=}')
                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)

                self.assertNotEqual(val, obj)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_datetime_obj(self):
        pifrl()
        from datetime import datetime
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: datetime = datetime.now()
                mapped_obj: datetime = creator.put_message(val)
                obj: datetime = creator.get_message()

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertEqual(mapped_obj, val)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('broken')
    def test_datetime_timedelta_obj(self):
        pifrl()
        from datetime import datetime, timedelta
        for _ in range(self.repeats_num):
            # TODO: Fix required
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: timedelta = datetime.now() - datetime.now()
                mapped_obj: timedelta = creator.put_message(val)
                obj: timedelta = creator.get_message()

                self.assertIsNot(val, mapped_obj)
                self.assertIsNot(val, obj)
                self.assertIsNot(mapped_obj, obj)

                self.assertEqual(mapped_obj, val)
                self.assertEqual(mapped_obj, obj)
    
    # @unittest.skip('temp skip')
    def test_dataclass_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = DataClass(1, 2)
                # print(f'{dir(val)=}')
                # pprint(val)
                # pprint(val.__dict__)
                mapped_val = creator.put_message(val)
                # print(f'{dir(mapped_val)=}')
                obj = creator.get_message()
                # print(f'{dir(obj)=}')
                obj = cast(DataClass, obj)
                obj.add_one()
                # pdi(val)
                # pdi(obj)
                # pdi(obj.__dict__)
                # print(f'{val.x=}, {obj.x=}')
                mapped_val.y = 'Yes'
                val.z = 40
                self.assertIsNot(val, obj)
                self.assertIs(val, mapped_val)
                self.assertIsNot(mapped_val, obj)

                self.assertNotEqual(val, obj)  # since internal __dict__ has no info about externally updated fields

                self.assertEqual(val.x, obj.x)
                self.assertEqual(val.y, obj.y)
                self.assertEqual(val.z, obj.z)
    
    # @unittest.skip('temp skip')
    def test_complex_dataclass_obj(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val: SomeSharedObject = SomeSharedObject(
                    some_processing_stage_control=False,
                    str_value='Hello, ',
                    data_dict={
                        'key1': 1,
                        ('key', 2): 'value2',
                        'key3': np.array([1, 2, 3], dtype=np.int32),
                    },
                    company_info=CompanyInfo(
                        company_id=1,
                        emails=('sails@company.com', 'support@company.com'),
                        websites=['http://company.com', 'http://company.org'],
                        income=1_000_000.0,
                        employees=10,
                        some_employee=Employee(
                            'John Doe', 
                            30, 
                            Positions.manager,
                            2,
                        )
                    )
                )

                # print(f'{dir(val)=}')
                # pprint(val)
                # pprint(val.__dict__)
                mapped_val = creator.put_message(val)
                # print(f'{dir(mapped_val)=}')
                obj = creator.get_message()

                obj = cast(SomeSharedObject, obj)
                obj.str_value += 'World!'
                obj.data_dict['key1'] *= 200
                obj.data_dict[('key', 2)] = obj.data_dict[('key', 2)] + ' data'
                obj.data_dict['key3'] += 10
                obj.company_info.income = 3_000_000.0
                obj.company_info.some_employee.increase_years_of_employment()
                obj.some_processing_stage_control = True

                self.assertIsNot(val, obj)
                self.assertIs(val, mapped_val)
                self.assertIsNot(mapped_val, obj)

                self.assertNotEqual(val, obj)  # since internal __dict__ has no info about externally updated fields

                self.assertEqual(val.str_value, obj.str_value)
                self.assertEqual(val.data_dict['key1'], obj.data_dict['key1'])
                self.assertEqual(val.data_dict[('key', 2)], obj.data_dict[('key', 2)])
                np.testing.assert_array_equal(val.data_dict['key3'], obj.data_dict['key3'])
                self.assertEqual(val.company_info.income, obj.company_info.income)
                self.assertEqual(val.company_info.some_employee.years_of_employment, obj.company_info.some_employee.years_of_employment)
                self.assertEqual(val.company_info.some_employee.age, obj.company_info.some_employee.age)
                self.assertEqual(val.some_processing_stage_control, obj.some_processing_stage_control)
    
    # @unittest.skip('temp skip')
    def test_mutable_set_rehash(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = set(range(1, 11))
                val2 = set(range(10, 101))
                val.update(val2)
                obj = creator.put_message(val)
                for item in val2:
                    obj.add(item)
                
                self.assertEqual(val, obj)
    
    # @unittest.skip('temp skip')
    def test_mutable_mapping_rehash(self):
        pifrl()
        for _ in range(self.repeats_num):
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                val = {i: i for i in range(1, 11)}
                val2 = {i: i for i in range(10, 101)}
                val.update(val2)
                obj = creator.put_message(val)
                obj.update(val2)
                
                self.assertEqual(val, obj)
    
    # @unittest.skip('temp skip')
    def test_shared_memory(self):
        pifrl()
        for _ in range(self.repeats_num):
            results: List[Any] = list()
            creator: SharedMemory = SharedMemory('test_shmem', True, 200 * 1024 * 1024)

            with creator:
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.malloc(ObjectType.tbool, 1)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.realloc(offset, 11)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                results.append(creator.free(offset))
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                creator.set_free_memory_search_start(0)
                creator.commit_free_memory_search_start()
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.malloc(ObjectType.tbool, 1)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                offset, size = creator.realloc(offset, 11)
                results.append(size)
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                results.append(creator.free(offset))
                results.append(creator.read_mem(creator.get_data_start_offset(), 50))
                creator.set_free_memory_search_start(0)
                creator.commit_free_memory_search_start()

                values = [
                    0,
                    1,
                    999999999999,
                    -1,
                    -999999999999,
                    0.0,
                    1.0,
                    999999999999.0,
                    -1.0,
                    -999999999999.0,
                    (1,),
                    [1],
                    (1, 2),
                    [1, 2],
                    {1, 2},
                    (1, [2, 3]),
                    {1: 1, 2: 2},
                    {1: 1, 2: [2, (3,)]},
                    'test',
                    'test'.encode('utf-8'),
                    bytearray('test'.encode('utf-8')),
                    True,
                    False,
                    {
                        'test': 1,
                        'test2': 2.0,
                        'test3': 'test',
                        'test4': [
                            1, 
                            2.0, 
                            True, 
                            False, 
                            'test', 
                            'test'.encode('utf-8'), 
                            bytearray('test'.encode('utf-8')), 
                            ('hello', 1),
                            ['world', 2.0],
                            {
                                'hello': [1, (2, 3)],
                                'world': 'world',
                            }
                        ],
                    }
                ]
                
                results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                results.append(creator.read_mem(creator.get_data_start_offset(), 500))
                
                for val in values:
                    creator.put_message(val)
                    results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                    results.append(creator.read_mem(creator.get_data_start_offset(), 500))
                    
                    message = creator.take_message()
                    self.assertEqual(val, message)
                    results.append(message)
                    results.append(creator.read_mem(creator.get_free_memory_search_start(), 100))
                    results.append(creator.read_mem(creator.get_data_start_offset(), 500))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


import unittest
from cengal.unittest.behavior_stabilizer import UnittestTestCaseState, UnittestTestCaseWithState
from cengal.math.numbers import ae
from cengal.time_management.run_time import RT
from cengal.introspection.inspect import get_exception

from cengal.parallel_execution.coroutines.coro_tools.run_in_loop import run_in_loop
from cengal.parallel_execution.coroutines.coro_scheduler import Interface, CoroWrapperBase, CoroID, GreenletExit
from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
from cengal.parallel_execution.coroutines.coro_standard_services.cpu_tick_count_per_second import CpuTickCountPerSecond
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro, PutCoroRequest
from cengal.parallel_execution.coroutines.coro_standard_services.kill_coro import KillCoro
from cengal.parallel_execution.coroutines.coro_standard_services.throw_coro import ThrowCoro
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_loop import ShutdownLoop
from cengal.parallel_execution.coroutines.coro_standard_services.shutdown_on_keyboard_interrupt import ShutdownOnKeyboardInterrupt
from cengal.parallel_execution.coroutines.coro_standard_services.db import *
from cengal.time_management.cpu_clock_cycles import perf_counter
from time import perf_counter
from typing import Tuple


async def await_for_cpu_tick_count(i: Interface) -> Tuple:
    cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation = await i(CpuTickCountPerSecond)
    while (last_ticks_per_second is None) or (val_99 < 1) or (val_95 < 1) or (val_68 < 1):
        cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation = await i(CpuTickCountPerSecond)
    
    return cpu_clock_cycles, last_ticks_per_second, val_99, val_95, val_68, max_deviation, min_deviation


class TestCaseForThrowCoro(unittest.TestCase):
    @unittest.skip('Skip')
    def test_kill_green(self):
        async def amy_coro(i: Interface, background_coro, a, b) -> int:
            try:
                await i(ShutdownOnKeyboardInterrupt)
                print('Started', perf_counter())
                print(await await_for_cpu_tick_count(i))
                coro_id: CoroID = await i(PutCoro, background_coro, 0.1)
                await i(Sleep, 0.5)
                await i(PutCoroRequest().put_background_coro(shutdown_loop, 5))
                await i(ThrowCoro, coro_id, MyException, MyException(), tree=True)
                print(await i(CpuTickCountPerSecond))
                return a + b
            except:
                ex = get_exception()
                print(ex)
            finally:
                print('Ended', perf_counter())

        with RT() as rt:
            result = run_in_loop(amy_coro)
        
        self.assertEqual(result, 7, 'Result is wrong')
        self.assertTrue(ae(rt.rt, 0.5, 0.5), f'Run time is wrong: {rt.rt}')
    
    # def test_kill_async(self):
    #     with RT() as rt:
    #         result = run_in_loop(amy_coro, abackground_coro, 2, 5)
        
    #     self.assertEqual(result, 7, 'Result is wrong')
    #     self.assertTrue(ae(rt.rt, 0.5, 0.5), f'Run time is wrong: {rt.rt}')


if __name__ == '__main__':
    unittest.main()
